from __future__ import annotations
from pathlib import Path
from typing import Optional
import difflib
import os

import librosa
import numpy as np
import pandas as pd
from torch.utils.data import Dataset
from torch.utils.data._utils.collate import default_collate_fn_map

from audidata.io.audio import load
from audidata.io.crops import RandomCrop
from audidata.transforms.audio import Mono
from audidata.transforms.midi import PianoRoll
from audidata.io.midi import read_single_track_midi
from audidata.collate.base import collate_list_fn


default_collate_fn_map.update({list: collate_list_fn})


class Isophonics(Dataset):
    r"""Isophonics [1] is a dataset contains a collection of albums maintained 
    by QMUL. The version until Dec. 2024 contains 283 songs from 18 CDs, with a 
    total duration of 15 hours. Audios are sampled at 44,100 Hz. After 
    decompression, the dataset is 687 MB. Some songs contains annotations of 
    beat, chord, structure annotated by QMUL researchers.

    Audios are from commercial CDs and are not publicly available. Users need
    to purchase the CDs (or ask for those who has the audio files). Annotations 
    can be downloaded from [1]. Annotations may contain errors. Please carefully
    read http://isophonics.net/content/reference-annotations for details.

    Confidence:

    | Dataset         | CDs | Songs | Beat confidence | Chord con.  | Key con. | Structure con. |
    |-----------------|-----|-------|-----------------|-------------|----------|----------------|
    | The Beatles     | 12  | 180   | Moderate        | High        | Low      | Good           |
    | Carole King     |  1  |  14   | -               | Not checked | Moderate | Good           |
    | Michael Jackson |  3  |  51   | -               | -           | -        | Good           |
    | Queen           |  2  |  38   | -               | Moderate    | Moderate | Good           |

    [1] http://isophonics.net/

    Users should prepare the dataset looks like:

        dataset_root (687 MB)
        ├── audio (283 mp3s)
        │   ├── 01_-_Please_Please_Me_01_-_I_Saw_Her_Standing_There.mp3
        │   └── ...
        └── annotations
            ├── beat
            │   └── The Beatles (12 CDs)
            ├── chordlab
            │   ├── Carole King (1 CD)
            │   ├── Queen (2 CDs)
            │   └── The Beatles (12 CDs)
            ├── keylab
            │   ├── Carole King (1 CD)
            │   ├── Queen (2 CDs)
            │   └── The Beatles (12 CDs)
            └── seglab
                ├── Carole King (1 CD)
                ├── Michael Jackson (2 CDs)
                ├── Queen (2 CDs)
                └── The Beatles (12 CDs)
    """

    url = "http://isophonics.net/"

    duration = 54184.58  # Dataset duration (s), 15 hours.

    def __init__(
        self, 
        root: str, 
        split: str = "train",
        sr: float = 16000,
        crop: Optional[callable] = RandomCrop(clip_duration=10., end_pad=9.9),
        transform: Optional[callable] = Mono(),
        target_types: bool = True,
        target_transform: Optional[callable] = PianoRoll(fps=100, pitches_num=128),
    ):

        self.root = root
        self.split = split
        self.sr = sr
        self.crop = crop
        self.target_types = target_types
        self.transform = transform
        self.target_transform = target_transform

        labels_dir = Path(self.root, "annotations", "chordlab")
        label_paths = sorted(list(Path(labels_dir).rglob('*.lab')))

        audios_dir = Path(self.root, "audios")
        audio_names = sorted(os.listdir(audios_dir))
        
        audio_paths = []

        from IPython import embed; embed(using=False); os._exit(0)
        for label_path in label_paths:

            name = str(Path(label_path.stem))
            print(name)
            audio_name = difflib.get_close_matches(name, audio_names)[0]
            audio_path = Path(audios_dir, audio_name)
            audio_paths.append(audio_path)

        meta_dict = {
            "audio_path": audio_paths,
            "label_path": label_paths
        }

        from IPython import embed; embed(using=False); os._exit(0)
            

        # meta_dict = {}

    def __getitem__(self, index: int) -> dict:

        from IPython import embed; embed(using=False); os._exit(0)

        audio_path = Path(self.root, self.meta_dict["audio_name"][index])
        midi_path = Path(self.root, self.meta_dict["midi_name"][index]) 
        duration = self.meta_dict["duration"][index]

        full_data = {
            "dataset_name": "MAESTRO-V3.0.0",
            "audio_path": str(audio_path),
        }

        # Load audio
        audio_data = self.load_audio(path=audio_path)
        full_data.update(audio_data)
        
        # Load target
        if self.target_types:
            target_data = self.load_target(
                midi_path=midi_path, 
                start_time=audio_data["start_time"],
                clip_duration=audio_data["duration"]
            )
            full_data.update(target_data)
        
        return full_data

    def __len__(self):

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def load_meta(self, meta_csv: str) -> dict:
        r"""Load meta dict.
        """

        df = pd.read_csv(meta_csv, sep=',')

        indexes = df["split"].values == self.split

        meta_dict = {
            "midi_name": df["midi_filename"].values[indexes],
            "audio_name": df["audio_filename"].values[indexes],
            "duration": df["duration"].values[indexes]
        }

        return meta_dict

    def load_audio(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)
        
        if self.crop:
            # Load a clip
            start_time, duration = self.crop(audio_duration=audio_duration)
        else:
            # Load full song
            start_time = 0.
            duration = audio_duration

        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=duration
        )
        # shape: (channels, audio_samples)

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": duration
        }

        if self.transform is not None:
            data = self.transform(data)

        return data

    def load_target(
        self, 
        midi_path: str, 
        start_time: float, 
        clip_duration: float
    ) -> dict:
        
        notes, pedals = read_single_track_midi(
            midi_path=midi_path, 
            extend_pedal=self.extend_pedal
        )
        
        data = {
            "note": notes,
            "pedal": pedals,
            "start_time": start_time,
            "clip_duration": clip_duration
        }
        
        if self.target_transform:
            data = self.target_transform(data)

        return data


from __future__ import annotations

import os
from pathlib import Path
from typing import Union

import librosa
import numpy as np
from audidata.io.audio import load
from audidata.io.crops import RandomCrop
from torch.utils.data import Dataset
from typing_extensions import Literal


class MUSDB18HQ(Dataset):
    r"""MUSDB18HQ [1] is a dataset containing 100 training audio files and 50 
    testing audio files, each with vocals, bass, drums, and other stems. The 
    total duration is 9.8 hours. The audio is stereo and sampled at 48,000 Hz. 
    After decompression, the dataset size is 30 GB.

    [1] https://zenodo.org/records/3338373

    The dataset looks like:

        musdb18hq (30 GB)
        ├── train (100 files)
        │   ├── A Classic Education - NightOwl
        │   │   ├── bass.wav
        │   │   ├── drums.wav
        │   │   ├── mixture.wav
        │   │   ├── other.wav
        │   │   └── vocals.wav
        │   ... 
        │   └── ...
        └── test (50 files)
            ├── Al James - Schoolboy Facination
            │   ├── bass.wav
            │   ├── drums.wav
            │   ├── mixture.wav
            │   ├── other.wav
            │   └── vocals.wav
            ... 
            └── ...
    """

    URL = "https://zenodo.org/records/3338373"

    DURATION = 35359.56  # Dataset duration (s), 9.8 hours, including training, 
    # valdiation, and testing

    def __init__(
        self,
        root: str = "/datasets/musdb18hq", 
        split: Literal["train", "test"] = "train",
        sr: int = 44100,
        crop: callable = RandomCrop(clip_duration=2.),
        target_stems: list[str] = ["vocals"],
        time_align: Literal["strict", "group", "random"] = "group",
        stem_transform: None | callable | list[callable] = None,
        group_transform: None | callable | list[callable] = None,
        mixture_transform: None | callable | list[callable] = None,
    ) -> None:
        r"""
        time_align: str. "strict" indicates all stems are aligned (from the 
            same song and have the same start time). "group" indictates 
            target stems / background stems are aligned. "random" indicates 
            all stems are from different songs with different start time.
        """

        self.stems = ["bass", "drums", "other", "vocals"]
        self.root = root
        self.split = split
        self.sr = sr
        self.crop = crop
        self.target_stems = target_stems
        self.bg_stems = list(set(self.stems) - set(self.target_stems))
        self.time_align = time_align
        self.stem_transform = group_transform
        self.group_transform = group_transform
        self.mixture_transform = group_transform

        if not Path(self.root).exists():
            raise Exception(f"{self.root} does not exist. Please download the dataset from {MUSDB18HQ.URL}")

        self.audios_dir = Path(self.root, self.split)
        self.list_names = sorted(os.listdir(self.audios_dir))
        self.audios_num = len(self.list_names)
        
    def __getitem__(
        self, 
        index: Union[int, dict],
    ) -> dict:

        # Use different song indexes for different stems
        index_dict = self.get_index_dict(index)
        # E.g., {"bass": 94, "drums": 94, "other": 35, "vocals": 35}

        audio_names = {}
        audio_paths = {}
        start_times = {}
        clip_durations = {}

        for stem in self.stems:
            
            audio_names[stem] = self.list_names[index_dict[stem]]
            audio_paths[stem] = Path(self.audios_dir, audio_names[stem], "{}.wav".format(stem))
            audio_duration = librosa.get_duration(path=audio_paths[stem])
            start_times[stem], clip_durations[stem] = self.crop(audio_duration=audio_duration)

        start_times = self.update_start_times(start_times)
        # E.g., {"bass": 44.86, "drums": 139.68, "other": 44.86, "vocals": 139.68}

        data = {
            "dataset_name": "MUSDB18HQ",
        }

        for stem in self.stems:

            # Load a clip
            data[stem] = load(
                path=audio_paths[stem], 
                sr=self.sr, 
                offset=start_times[stem], 
                duration=clip_durations[stem],
            )
            # shape: (channels, audio_samples)

            data["{}_audio_name".format(stem)] = audio_names[stem]
            data["{}_start_time".format(stem)] = start_times[stem]

            # Transform source
            if self.stem_transform is not None:
                data[stem] = self.stem_transform(data[stem])

        # Sum sources to target and background
        data["target"], data["background"] = self.sources_to_target_and_background(data)

        # Transform target and background
        if self.group_transform is not None:
            data["target"] = self.group_transform(data["target"])
            data["background"] = self.group_transform(data["background"])

        # Sum target and background to mixture
        data["mixture"] = data["target"] + data["background"]

        # Transform mixture
        if self.mixture_transform is not None:
            data["mixture"] = self.mixture_transform["mixture"]

        return data

    def __len__(self) -> int:
        return self.audios_num

    def get_index_dict(self, index: Union[int, dict]) -> dict:
        r"""Get song indexes of different stems."""

        if isinstance(index, int):
            # All sources have same indexes (from the same song)
            index_dict = {stem: index for stem in self.stems}

        elif isinstance(index, dict):
            index_dict = index

            if self.time_align == "strict":
                assert(self.equal_values(index_dict.values()))

            elif self.time_align == "group":
                assert(self.equal_values([index_dict[stem] for stem in self.target_stems]))
                assert(self.equal_values([index_dict[stem] for stem in self.bg_stems]))
                
            elif self.time_align == "random":
                pass

        else:
            raise TypeError(index)

        return index_dict

    def equal_values(self, x: list) -> bool:
        r"""Check if all values are the same."""
        return len(set(x)) == 1

    def update_start_times(self, start_times: dict) -> dict:
        r"""Update start times according to different time_align types."""

        if self.time_align == "strict":
            for stem in self.stems:
                start_times[stem] = start_times[self.stems[0]]
        
        elif self.time_align == "group":
            
            for stem in self.target_stems:
                start_times[stem] = start_times[self.target_stems[0]]
            
            for stem in self.bg_stems:
                start_times[stem] = start_times[self.bg_stems[0]]

        elif self.time_align == "random":
            pass

        else:
            raise ValueError(self.time_align)

        return start_times

    def sources_to_target_and_background(self, data: dict) -> tuple[np.ndarray, np.ndarray]:
        r"""Sum sources to target and background."""

        target = 0
        bg = 0

        for stem in self.target_stems:
            target += data[stem]

        for stem in self.bg_stems:
            bg += data[stem]

        return target, bg


from __future__ import annotations

from pathlib import Path

import librosa
import pandas as pd
from audidata.io.audio import load
from audidata.io.crops import StartCrop
from audidata.transforms.audio import Mono
from audidata.utils import call
from torch.utils.data import Dataset
from typing_extensions import Literal


class Clotho(Dataset):
    r"""Clotho [1] is an audio caption dataset consisting of 4,981 audio 
    samples, each with 5 captions. The total duration is 24.5 hours. The audio 
    samples range from 15 to 30 seconds in length. The captions are 8 to 20 
    words long. All audio are mono sampled at 44,100 Hz. After decompression, 
    the dataset size is 7.3 GB.

    [1] K. Drossos, et al., Clotho: An Audio Captioning Dataset, ICASSP, 2020

    The dataset looks like:

        clotho (7.3 GB)
        ├── clotho_audio_development (2894 wavs)
        ├── clotho_audio_evaluation (1046 wavs)
        ├── clotho_captions_development.csv
        ├── clotho_captions_evaluation.csv
        ├── clotho_metadata_development.csv
        ├── clotho_metadata_evaluation.csv
        └── LICENSE
    """

    URL = "https://zenodo.org/records/3490684"

    DURATION = 88366.30  # Dataset duration (s), 24.5 hours, including 
    # development and evaluation.

    def __init__(
        self, 
        root: str = None, 
        split: Literal["train", "test"] = "train",
        sr: float = 44100,  # Sampling rate
        crop: None | callable = StartCrop(clip_duration=10.),
        transform: None | callable = Mono(),
        target_transform: None | callable = None
    ) -> None:
    
        self.root = root
        self.split = split
        self.sr = sr
        self.crop = crop
        self.transform = transform
        self.target_transform = target_transform

        if self.split == "train":
            self.meta_csv = Path(self.root, "clotho_captions_development.csv")
            self.audios_dir = Path(self.root, "clotho_audio_development")

        elif self.split == "test":
            self.meta_csv = Path(self.root, "clotho_captions_evaluation.csv")
            self.audios_dir = Path(self.root, "clotho_audio_evaluation")

        else:
            raise ValueError(self.split)

        self.meta_dict = self.load_meta(self.meta_csv)

        if not Path(self.root).exists():
            raise Exception(f"{self.root} does not exist. Please download the dataset from {Clotho.URL}")

    def __getitem__(self, index: int) -> dict:

        
        audio_path = self.meta_dict["audio_path"][index]
        caption = self.meta_dict["caption"][index]

        full_data = {
            "dataset_name": "Clotho",
            "audio_path": audio_path,
        }

        # Load audio data
        audio_data = self.load_audio_data(path=audio_path)
        full_data.update(audio_data)

        # Load target data
        target_data = self.load_target_data(caption=caption)
        full_data.update(target_data)

        return full_data

    def __len__(self) -> int:

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def load_meta(self, meta_csv: str) -> dict:

        meta_dict = {"audio_name": [], "audio_path": [], "caption": []}

        df = pd.read_csv(meta_csv, sep=',')

        for n in range(len(df)):
            for i in range(1, 6):

                audio_name = df["file_name"][n]
                audio_path = str(Path(self.audios_dir, audio_name))

                meta_dict["audio_name"].append(audio_name)
                meta_dict["audio_path"].append(audio_path)
                meta_dict["caption"].append(df["caption_{}".format(i)][n])

        return meta_dict

    def load_audio_data(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)
        
        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            clip_duration = audio_duration

        # Load a clip
        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )  # shape: (channels, audio_samples)

        # Transform audio
        if self.transform is not None:
            audio = call(transform=self.transform, x=audio)

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration
        }

        return data

    def load_target_data(self, caption: str) -> dict:

        target = caption

        # Transform target
        if self.target_transform:
            target = call(transform=self.target_transform, x=target)

        data = {
            "caption": caption,
            "target": target,
        }

        return data


from __future__ import annotations

from pathlib import Path

import librosa
import pandas as pd
from audidata.io.audio import load
from audidata.io.crops import StartCrop
from audidata.transforms.audio import Mono
from audidata.utils import call
from torch.utils.data import Dataset
from typing_extensions import Literal


class ClothoAQA(Dataset):
    r"""Clotho-AQA [1] is an audio query-answering dataset consisting of 1,991
    audio samples from the Clotho dataset. Each audio sample has 6 associated 
    questions. For each question, the answers are provided by three different 
    annotators making a total of 35,838 question-answer pairs. The total 
    duration is 12.5 hours. The audio samples range from 15 to 30 seconds in 
    length. After decompression, the dataset size is 3.7 GB.

    [1] S. Lipping, et al., Clotho-AQA: A Crowdsourced Dataset for Audio 
    Question Answering, ArXiv, 2020

    The dataset looks like:

        clotho-aqa (3.7 GB)
        ├── audio_files (2894 wavs)
        ├── clotho_aqa_metadata.csv
        ├── clotho_aqa_test.csv
        ├── clotho_aqa_train.csv
        ├── clotho_aqa_val.csv
        └── LICENSE.txt
    """

    URL = "https://zenodo.org/records/6473207"

    DURATION = 44847.45  # Dataset duration (s), 12.5 hours, including 
    # train, val, and test sets.

    def __init__(
        self, 
        root: str = None, 
        split: Literal["train", "val", "test"] = "train",
        sr: float = 44100,  # Sampling rate
        crop: None | callable = StartCrop(clip_duration=10.),
        transform: None | callable = Mono(),
        question_transform: None | callable = None,
        target_transform: None | callable = None
    ) -> None:
    
        self.root = root
        self.split = split
        self.sr = sr
        self.crop = crop
        self.transform = transform
        self.question_transform = question_transform
        self.target_transform = target_transform

        if not Path(self.root).exists():
            raise Exception(f"{self.root} does not exist. Please download the dataset from {Clotho.URL}")

        self.audios_dir = Path(root, "audio_files")
        self.meta_csv = Path(root, "clotho_aqa_{}.csv".format(split))

        self.meta_dict = self.load_meta(self.meta_csv)

    def __getitem__(self, index: int) -> dict:

        
        audio_path = self.meta_dict["audio_path"][index]
        answering = self.meta_dict["answering"][index]
        question = self.meta_dict["question"][index]

        full_data = {
            "dataset_name": "Clotho-AQA",
            "audio_path": audio_path,
        }

        # Load audio data
        audio_data = self.load_audio_data(path=audio_path)
        full_data.update(audio_data)

        # Load question data
        question_data = self.load_question_data(question=question)
        full_data.update(question_data)

        # Load target data
        target_data = self.load_target_data(answering=answering)
        full_data.update(target_data)

        return full_data

    def __len__(self) -> int:

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def load_meta(self, meta_csv: str) -> dict:

        meta_dict = {
            "audio_name": [], 
            "audio_path": [], 
            "question": [], 
            "answering": []
        }

        df = pd.read_csv(meta_csv, sep=',')

        for n in range(len(df)):

            audio_name = df["file_name"][n]
            audio_path = str(Path(self.audios_dir, audio_name))

            meta_dict["audio_name"].append(audio_name)
            meta_dict["audio_path"].append(audio_path)
            meta_dict["question"].append(df["QuestionText"][n])
            meta_dict["answering"].append(df["answer"][n])

        return meta_dict

    def load_audio_data(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)
        
        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            clip_duration = audio_duration

        # Load a clip
        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )  # shape: (channels, audio_samples)

        # Transform audio
        if self.transform is not None:
            audio = call(transform=self.transform, x=audio)

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration
        }

        return data

    def load_question_data(self, question: str) -> dict:

        # Transform target
        if self.target_transform:
            question = call(transform=self.target_transform, x=question)

        data = {
            "question": question
        }

        return data

    def load_target_data(self, answering: str) -> dict:

        target = answering

        # Transform target
        if self.target_transform:
            target = call(transform=self.target_transform, x=target)

        data = {
            "answering": answering,
            "target": target,
        }

        return data


from __future__ import annotations
import random
from typing import Callable
import numpy as np


def call(
    transform: callable | list[callable], 
    x: Any
) -> Any:
    r"""Transform input/output data."""

    if isinstance(transform, Callable):
        return transform(x)

    elif isinstance(transform, list):
        for trans in transform:
            x = trans(x)
        return x

    else:
        raise TypeError(transform)


class RandomChoice:
    def __init__(self, callables: callable, weights: list[float]):
        
        self.callables = callables
        self.weights = weights

    def __call__(self, **kwargs) -> Any:

        call = random.choices(
            population=self.callables, 
            weights=self.weights
        )[0]

        return call(**kwargs)



from __future__ import annotations

import re


class TextNormalization:
    def __init__(self) -> None:
        pass

    def __call__(self, x: str) -> str:
        r"""Normalize a string. 

		E.g., "How are you? Fine, thank you!" => "how are you fine thank you"

        From: https://github.com/XinhaoMei/WavCaps/blob/master/captioning/data_handling/text_transform.py
        """

        # Transform to lower case
        x = x.lower()

        # Remove any forgotten space before punctuation and double space
        x = re.sub(r'\s([,.!?;:"](?:\s|$))', r'\1', x).replace('  ', ' ')

        # Remove punctuations
        x = re.sub('[(,.!?;:|*\")]', ' ', x).replace('  ', ' ')

        return x


import bisect
import itertools
import re
from typing import Any, Union

import numpy as np


class BaseTokenizer:
    r"""Base class for all tokenizers.
    """

    def __init__(self, words: list[Any]):
        self.words = words
        self.vocab_size = len(self.words)

        self.token_to_word = {token: word for token, word in enumerate(self.words)}
        self.word_to_token = {word: token for token, word in enumerate(self.words)}

    def stoi(self, word: Any) -> int:
        r"""String (word) to index.
        """
        return self.word_to_token.get(word)
        
    def itos(self, token: int) -> Any:
        r"""Index to string (word).
        """
        assert 0 <= token < self.vocab_size
        return self.token_to_word[token]


class SpecialTokenizer(BaseTokenizer):
    def __init__(self):
        words = ["<pad>", "<sos>", "<eos>", "<unk>"]
        super().__init__(words=words)
        

class NameTokenizer(BaseTokenizer):
    def __init__(self):
        words = [
            "note_on", "note_off", "note_sustain",
            "pedal_on", "pedal_off", "pedal_sustain",
            "beat", "downbeat",
        ]
        words = ["name={}".format(w) for w in words]

        vocab_size = 16  # Reserve for future
        words = pad_list(words, vocab_size)

        super().__init__(words=words)


def pad_list(x: list, max_len: int) -> list:
    assert len(x) <= max_len
    while len(x) < max_len:
        x.append("blank_{}".format(len(x)))
    return x


class TimeTokenizer(BaseTokenizer):
    def __init__(self, max_duration=60., fps=100):

        self.fps = fps
        self.delta_t = 1. / fps

        words = ["time={:.4f}".format(time) for time in \
            np.arange(0, max_duration + self.delta_t, self.delta_t)]

        super().__init__(words=words) 


    def stoi(self, word: str) -> int:

        if "time=" in word:
            time = float(re.search('time=(.*)', word).group(1))
            time = round(time * self.fps) / self.fps
            word = "time={:.4f}".format(time)
            token = super().stoi(word)
            return token
    

class PitchTokenizer(BaseTokenizer):
    def __init__(self, classes_num=128):

        words = ["pitch={}".format(pitch) for pitch in range(classes_num)]

        super().__init__(words=words)

    
class VelocityTokenizer(BaseTokenizer):
    def __init__(self, classes_num=128):
        
        words = ["velocity={}".format(pitch) for pitch in range(classes_num)]

        super().__init__(words=words)
        
class DrumTokenizer(BaseTokenizer):
    def __init__(self, classes_num=128):
        
        words = ["drum_pitch={}".format(pitch) for pitch in range(classes_num)]

        super().__init__(words=words)
        

class ConcatTokenizer:
    def __init__(self, tokenizers, verbose=False):

        self.tokenizers = tokenizers

        self.words = list(itertools.chain(*[tk.words for tk in tokenizers]))
        self.vocab_sizes = [tk.vocab_size for tk in self.tokenizers]
        self.vocab_size = np.sum(self.vocab_sizes)

        self.cumulative_sizes = np.cumsum(self.vocab_sizes)

        if verbose:
            print("Vocab size: {}".format(self.vocab_size))
            for tk in self.tokenizers:
                print(tk.vocab_size)

    def stoi(self, word: str) -> int:
        
        start_token = 0

        for tk in self.tokenizers:
            
            token = tk.stoi(word)
            
            if token is not None:
                return start_token + token
            else:
                start_token += tk.vocab_size

        raise NotImplementedError("{} is not in the vocabulary!".format(word))

    def itos(self, token: int) -> str:
        
        assert 0 <= token < self.vocab_size

        tokenizer_idx = bisect.bisect_right(self.cumulative_sizes, token)
        tokenizer = self.tokenizers[tokenizer_idx]
        
        if tokenizer_idx == 0:
            rel_token = token
        else:
            rel_token = token - self.cumulative_sizes[tokenizer_idx - 1]
        
        word = tokenizer.itos(rel_token)

        return word
    
class ProgramTokenizer(BaseTokenizer):
    def __init__(self):
        words = [f"program={i}" for i in range(128)]
        super().__init__(words=words)

    def stoi(self, word: Union[str, int]) -> int:
        if isinstance(word, int):
            if 0 <= word < 128:
                return word
        elif isinstance(word, str) and word.startswith("program="):
            try:
                program = int(word.split("=")[1])
                if 0 <= program < 128:
                    return program
            except ValueError:
                pass
        return self.word_to_token.get("<unk>", 0)  # Default to 0 (Piano). TODO: discuss

    def itos(self, token: int) -> str:
        if 0 <= token < 128:
            return f"program={token}"
        return "program=0"
    
class DictTokenizer(BaseTokenizer):
    def __init__(self, key_tokenizer_pairs):
        self.key_tokenizer_pairs = key_tokenizer_pairs
        self.keys = list(key_tokenizer_pairs.keys())
        self.tokenizers = list(key_tokenizer_pairs.values())
        
        words = ["<bot>", "<eot>"]  # begin of token, end of token
        for key in self.keys:
            words.append(f"<{key}>")  # Add a token for each key
        for tokenizer in self.tokenizers:
            for word in tokenizer.words:
                words.append(f"<{word}>")
        
        super().__init__(words=words)
        
        self.bot_token = self.word_to_token["<bot>"]
        self.eot_token = self.word_to_token["<eot>"]
        self.key_tokens = {key: self.word_to_token[f"<{key}>"] for key in self.keys}
        self.vocab_size = len(self.words)
    
    def tokenize(self, sequence):
        tokens = []
        current_dict = None
        for item in sequence:
            if item == "<bot>":
                if current_dict is not None:
                    tokens.append(current_dict)
                current_dict = {key: self.key_tokens[key] for key in self.keys}  
                # Initialize with key tokens, therefore if a key has not been used, it will be initialized with the key token.
            elif item == "<eot>":
                if current_dict is not None:
                    tokens.append(current_dict)
                    current_dict = None
            else:
                key, value = item[1:-1].split(":", 1)
                tokenizer = self.key_tokenizer_pairs[key]
                token = tokenizer.stoi(value)
                if token is not None and current_dict is not None:
                    current_dict[key] = token
        
        if current_dict is not None:
            tokens.append(current_dict)
        
        return tokens
    
    def detokenize(self, tokens):
        sequence = []
        for token_dict in tokens:
            sequence.append("<bot>")
            for key, value in token_dict.items():
                if value == self.key_tokens[key]:
                    continue  # Skip if it's just the key token
                tokenizer = self.key_tokenizer_pairs[key]
                word = tokenizer.itos(value)
                sequence.append(f"<{key}:{word}>")
            sequence.append("<eot>")
        
        return sequence
    
    def get_vocab_sizes(self):
        vocab_dict = {}
        for key, tokenizer in self.key_tokenizer_pairs.items():
            vocab_dict[key] = tokenizer.vocab_size + 1 # Add 1 for the key token
        return vocab_dict


if __name__ == '__main__':
    r"""Example.
    """
    
    print("Testing ConcatTokenizer")

    tokenizer = ConcatTokenizer([
        SpecialTokenizer(),
        NameTokenizer(),
        TimeTokenizer(),
        PitchTokenizer(),
        DrumTokenizer(),
        VelocityTokenizer(),
        ProgramTokenizer(),
    ])

    token = tokenizer.stoi("name=note_on")
    word = tokenizer.itos(token)
    print(token, word)

    token = tokenizer.stoi("time=26.789")
    word = tokenizer.itos(token)
    print(token, word)

    token = tokenizer.stoi("pitch=34")
    word = tokenizer.itos(token)
    print(token, word)

    token = tokenizer.stoi("velocity=34")
    word = tokenizer.itos(token)
    print(token, word)

    token = tokenizer.stoi("program=34")
    word = tokenizer.itos(token)
    print(token, word)
    
    token = tokenizer.stoi("drum_pitch=34")
    word = tokenizer.itos(token)
    print(token, word)
    
    print("Testing DictTokenizer")
    
    tokenizer = DictTokenizer({
        "special": SpecialTokenizer(),
        "onset": ConcatTokenizer([
            TimeTokenizer(),
            NameTokenizer(),
        ]),
        "pitch": PitchTokenizer(),
        "velocity": VelocityTokenizer(),
        "offset": ConcatTokenizer([
            TimeTokenizer(),
            NameTokenizer(),
        ]),
    })
    
    print("Vocab sizes:", tokenizer.get_vocab_sizes())

    new_sequence = ["<bot>", "<special:<sos>>", "<eot>", "<bot>", "<onset:time=1.9000>", "<pitch:pitch=34>", "<velocity:velocity=34>", "<offset:time=1.9700>", "<eot>"]
    new_tokens = tokenizer.tokenize(new_sequence)
    print("New sequence tokens:", len(new_tokens), "tokens, ", len(new_sequence), "words, ", new_tokens)
    new_text = tokenizer.detokenize(new_tokens)
    print("New sequence detokenized:", new_text)


from __future__ import annotations
import os
import random
from pathlib import Path
from typing import Optional
import pandas as pd
import torch
import librosa
import numpy as np
from torch.utils.data import Dataset

from audidata.io.audio import load
from audidata.io.crops import StartCrop
from audidata.transforms.audio import Mono


class MagnaTagATune(Dataset):
    r"""MagnaTagATune[1] dataset contains 25,863 music clips.
    Each clip is a 29-seconds-long excerpt belonging to one of the 5223 songs, 445 albums and 230 artists.
    The clips span a broad range of genres like Classical, New Age, Electronica, Rock, Pop, World, Jazz, Blues, Metal, Punk, and more.
    Each audio clip is supplied with a vector of binary annotations of 188 tags.
    There are 16 parts, and researchers comonnly use parts 1-12 for training, part 13 for validation and parts 14-16 for testing.

    [1] Law, E., West, K., Mandel, M. I., Bay, M., and Downie, J. S. (2009). Evaluation of algorithms
    using games: The case of music tagging. In ISMIR, pages 387–392. Citeseer.

    """

    url = "https://mirg.city.ac.uk/codeapps/the-magnatagatune-dataset"

    duration = 75708000  # Dataset duration (s), 21,050 hours, including 
    # training, valdiation, and testing

    top_50 = [
        "guitar", "classical", "slow", "techno", "strings", "drums", "electronic", "rock", "fast", "piano",
        "ambient", "beat", "violin", "vocal", "synth", "female", "indian", "opera", "male", "singing", "vocals",
        "no vocals", "harpsichord", "loud", "quiet", "flute", "woman", "male vocal", "no vocal", "pop", "soft",
        "sitar", "solo", "man", "classic", "choir", "voice", "new age", "dance", "male voice", "female vocal",
        "beats", "harp", "cello", "no voice", "weird", "country", "metal", "female voice", "choral"
    ]

    classes_num = len(top_50)
    lb_to_ix = {lb: ix for ix, lb in enumerate(top_50)}
    ix_to_lb = {ix: lb for ix, lb in enumerate(top_50)}

    def __init__(
            self,
            root: str = None,
            split: ["train", "test", "val"] = "train",
            sr: int = 16000,
            crop: Optional[callable] = StartCrop(clip_duration=29.),
            transform: Optional[callable] = Mono(),
            target_transform: Optional[callable] = None
    ):

        self.root = root
        self.split = split
        self.sr = sr
        self.crop = crop
        self.transform = transform
        self.target_transform = target_transform

        meta_csv = Path(self.root, "annotations_final.csv")
        self.meta_dict = self.load_meta(meta_csv)

    def __getitem__(self, index: int) -> dict:

        audio_path = self.meta_dict["audio_path"][index]
        labels = self.meta_dict["labels"][index]

        full_data = {
            "dataset_name": "MagnaTagATune",
            "audio_path": str(audio_path),
        }

        # Load audio
        audio_data = self.load_audio(path=audio_path)
        full_data.update(audio_data)

        # Load target
        target_data = self.load_target(labels=labels)
        full_data.update(target_data)

        return full_data

    def __len__(self) -> int:

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def load_meta(self, meta_csv) -> dict:
        r"""Load metadata of the MagnaTagATune dataset."""
        top_50 = MagnaTagATune.top_50
        meta_dict = {
            "labels": [],
            "audio_name": [],
            "audio_path": []
        }

        df = pd.read_csv(meta_csv, sep="\t")
        df = df[df[top_50].sum(axis=1) > 0]
        # remove the columns that are not in top_50 + mp3_path and clip_id
        df = df[top_50 + ["mp3_path", "clip_id"]]

        # Load train, validation, and test IDs
        if self.split == "train":
            ids_df = pd.read_csv(os.path.join(self.root, "train_gt_mtt.tsv"), sep="\t", header=None)
        elif self.split == "val":
            ids_df = pd.read_csv(os.path.join(self.root, "val_gt_mtt.tsv"), sep="\t", header=None)
        elif self.split == "test":
            ids_df = pd.read_csv(os.path.join(self.root, "test_gt_mtt.tsv"), sep="\t", header=None)

        ids = ids_df[0].tolist()

        split_df = df[df["clip_id"].isin(ids)]

        label_names = df.columns.drop(["mp3_path", "clip_id"])

        for index, row in split_df.iterrows():
            # Get the mp3_path
            audio_name = row["mp3_path"]
            audio_path = os.path.join(self.root, audio_name)

            # Generate list of labels based on the row values
            labels = [label for label in label_names if row[label] == 1]

            # If the song has no tags, skip it
            if not labels:
                continue

            meta_dict["labels"].append(labels)
            meta_dict["audio_name"].append(audio_name)
            meta_dict["audio_path"].append(audio_path)

        return meta_dict

    def load_audio(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)

        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            duration = None

        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )
        # shape: (channels, audio_samples)

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration if clip_duration else audio_duration
        }

        if self.transform is not None:
            data = self.transform(data)

        return data

    def load_target(self, labels: list[str]) -> dict:
        classes_num = MagnaTagATune.classes_num
        lb_to_ix = MagnaTagATune.lb_to_ix

        target = np.zeros(classes_num, dtype="float32")

        target[[lb_to_ix[label] for label in labels]] = 1

        data = {
            "target": target,
            "labels": labels
        }

        return data


import random

from torch.utils.data import Dataset


class InfiniteSampler:
    r"""Randomly yield index infinitely.
    """

    def __init__(self, dataset: Dataset):

        self.indexes = list(range(len(dataset)))
        random.shuffle(self.indexes)
        
    def __iter__(self) -> int:

        pointer = 0

        while True:

            if pointer == len(self.indexes):
                random.shuffle(self.indexes)
                pointer = 0
                
            index = self.indexes[pointer]
            pointer += 1

            yield index


from __future__ import annotations

from pathlib import Path
from typing import NoReturn

import librosa
import pandas as pd
from audidata.io.audio import load
from audidata.io.crops import StartCrop
from audidata.transforms.audio import Mono
from audidata.utils import call
from torch.utils.data import Dataset
from typing_extensions import Literal


class AudioCaps(Dataset):
    r"""AudioCaps [1] is an audio caption dataset containing 51,308 audio clips, 
    each accompanied by one caption. The audio samples are approximately 10 
    seconds long and are mono sampled at 32,000 Hz. After decompression, the 
    dataset size is 131 GB.

    [1] C. D. Kim, et al. AudioCaps: Generating Captions for Audios in The Wild, NAACL-HLT 2019

    The dataset looks like:

        audiocaps (131 GB)
        ├── train (49274 files)
        ├── val (494 files)
        ├── test (957 files)
        ├── train.csv
        ├── val.csv
        ├── test.csv
        ├── LICENSE.txt
        └── README.md
    """

    URL = "https://github.com/cdjkim/audiocaps"
    
    DURATION = 501667.11  # Dataset duration (s), 139 hours, including training, 
    # validation, and testing

    def __init__(
        self, 
        root: str = None, 
        split: Literal["train", "val" "test"] = "train",
        sr: float = 32000,  # Sampling rate
        crop: None | callable = StartCrop(clip_duration=10.),
        transform: None | callable = Mono(),
        target_transform: None | callable = None
    ) -> NoReturn:
    
        self.root = root
        self.split = split
        self.sr = sr
        self.crop = crop
        self.transform = transform
        self.target_transform = target_transform

        self.meta_csv = Path(self.root, "{}.csv".format(self.split))
        self.audios_dir = Path(self.root, self.split)

        self.meta_dict = self.load_meta(self.meta_csv)

        if not Path(self.root).exists():
            raise Exception(f"{self.root} does not exist. Please download the dataset from {AudioCaps.URL}")

    def __getitem__(self, index: int) -> dict:

        audio_path = self.meta_dict["audio_path"][index]
        caption = self.meta_dict["caption"][index]
        
        full_data = {
            "dataset_name": "AudioCaps",
            "audio_path": audio_path,
        }

        # Load audio data
        audio_data = self.load_audio_data(path=audio_path)
        full_data.update(audio_data)

        # Load target data
        target_data = self.load_target_data(caption=caption)
        full_data.update(target_data)

        return full_data

    def __len__(self) -> int:

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def load_meta(self, meta_csv: str) -> dict:

        df = pd.read_csv(meta_csv, sep=',')

        meta_dict = {
            "audio_name": [], 
            "audio_path": [],
            "audiocap_id": [], 
            "caption": []
        }

        for n in range(len(df)):
            
            audio_name = "Y{}.wav".format(df["youtube_id"][n])
            audio_path = str(Path(self.audios_dir, audio_name))

            meta_dict["audio_name"].append(audio_name)
            meta_dict["audio_path"].append(audio_path)
            meta_dict["audiocap_id"].append(df["audiocap_id"][n])
            meta_dict["caption"].append(df["caption"][n])

        return meta_dict

    def load_audio_data(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)

        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            clip_duration = audio_duration

        # Load a clip
        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )  # shape: (channels_num, audio_samples)

        # Transform audio
        if self.transform is not None:
            audio = call(transform=self.transform, x=audio)

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration
        }

        return data

    def load_target_data(self, caption: str) -> dict:

        target = caption

        # Transform target
        if self.target_transform:
            target = call(transform=self.target_transform, x=target)

        data = {
            "caption": caption,
            "target": target,
        }

        return data


import numpy as np


class OneHot:
    r"""Convert label to one-hot representation.
    """

    def __init__(self, classes_num: int):
        self.classes_num = classes_num

    def __call__(self, index: int) -> np.ndarray:
        
        target = np.zeros(self.classes_num, dtype="float32")
        target[index] = 1

        return target


from __future__ import annotations
import os
from pathlib import Path
from typing import Optional
from typing_extensions import Literal
import pandas as pd
import librosa
import numpy as np
from torch.utils.data import Dataset

from audidata.io.audio import load
from audidata.io.crops import StartCrop
from audidata.transforms.audio import Mono


class ESC50(Dataset):
    r"""The ESC-50 [1] dataset is a labeled collection of 2000 environmental audio recordings suitable for benchmarking methods of environmental sound classification.
    It comprises 2000 5s-clips of 50 different classes across natural, human and domestic sounds, again, drawn from Freesound.org.

    [1] Piczak, Karol J. "ESC: Dataset for environmental sound classification." Proceedings of the 23rd ACM international conference on Multimedia. 2015.
    """

    url = "https://github.com/karolpiczak/ESC-50?tab=readme-ov-file#download"

    duration = 10008  # Dataset duration (s), 21,050 hours, including

    def __init__(
            self,
            root: str = None,
            sr: int = 44100,
            fold: Optional[int] = None,
            split: Optional[Literal["train", "test"]] = None,
            crop: Optional[callable] = StartCrop(clip_duration=29.),
            transform: Optional[callable] = Mono(),
            target_transform: Optional[callable] = None
    ):

        self.root = root
        self.sr = sr
        assert (fold is None) == (split is None)
        self.fold = fold  # the current fold for the 5-fold cross-validation in the ESC-50 dataset.
        self.split = split
        self.crop = crop
        self.transform = transform
        self.target_transform = target_transform

        self.meta_dict = self.load_meta()

    def __getitem__(self, index: int) -> dict:

        audio_path = self.meta_dict["audio_path"][index]
        label_id = self.meta_dict["label_id"][index]
        label = self.meta_dict["label"][index]

        target = np.zeros(self.classes_num, dtype="float32")
        target[label_id] = 1

        full_data = {
            "dataset_name": "ESC50",
            "audio_path": str(audio_path),
            "label": label,
            "target": target
        }

        # Load audio
        audio_data = self.load_audio(path=audio_path)
        full_data.update(audio_data)

        return full_data

    def __len__(self) -> int:

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def load_meta(self) -> dict:
        r"""Load metadata of the ESC-50 dataset."""
        meta_csv = os.path.join(self.root, 'meta', 'esc50.csv')
        metadata = pd.read_csv(meta_csv)

        self.classes_num = len(metadata['target'].unique())

        meta_dict = {
            "audio_path": [],
            "audio_name": [],
            "label_id": [],
            "label": []
        }

        for _, row in metadata.iterrows():
            fold = row['fold'] - 1
            if self.fold is not None and self.split == 'train' and self.fold == fold:
                continue
            if self.fold is not None and self.split == 'test' and self.fold != fold:
                continue
            audio_path = os.path.join(self.root, 'audio', row['filename'])
            audio_name = row['filename']
            label_id = row['target']
            label = row['category']

            meta_dict["audio_name"].append(audio_name)
            meta_dict["audio_path"].append(audio_path)
            meta_dict["label_id"].append(label_id)
            meta_dict["label"].append(label)

        return meta_dict

    def load_audio(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)

        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            duration = None

        audio = load(
            path=path,
            sr=self.sr,
            offset=start_time,
            duration=clip_duration
        )
        # shape: (channels, audio_samples)

        data = {
            "audio": audio,
            "start_time": start_time,
            "duration": clip_duration if clip_duration else audio_duration
        }

        if self.transform is not None:
            data = self.transform(data)

        return data


from __future__ import annotations
import random


class StartCrop:
    r"""Prepare start time and duration of to crop from the start.
    """

    def __init__(self, clip_duration: float):
        self.start_time = 0.
        self.clip_duration = clip_duration

    def __call__(self, **kwargs) -> tuple[float, float]:
        return self.start_time, self.clip_duration


class RandomCrop:
    r"""Prepare start time and duration of to crop from random time.
    """

    def __init__(
        self, 
        clip_duration: float, 
        end_pad: float = 0  # Pad silent at the end (s)
    ):
        self.clip_duration = clip_duration
        self.end_pad = end_pad

    def __call__(self, audio_duration: float) -> tuple[float, float]:

        padded_duration = audio_duration + self.end_pad

        if self.clip_duration <= padded_duration:
            start_time = random.uniform(0., padded_duration - self.clip_duration)

        else:
            start_time = 0

        return start_time, self.clip_duration


from __future__ import annotations

from pathlib import Path

import librosa
import numpy as np
import pandas as pd
from audidata.io.audio import load
from audidata.io.crops import StartCrop
from audidata.transforms.audio import Mono
from audidata.utils import call
from torch.utils.data import Dataset
from typing_extensions import Literal


class LJSpeech(Dataset):
    r"""LJSpeech [1] is a speech dataset containing 13,100 audio clips from a 
    single speaker. The total duration is 24 hours of speech. The audio clips 
    range in duration from 1.11 seconds to 10.10 seconds. All audio are mono 
    sampled at 22,050 Hz. The captions range from 1 to 59 words. After 
    decompression, the dataset size is 3.6 GB.

    [1] https://keithito.com/LJ-Speech-Dataset/

    The dataset looks like:

        LJSpeech-1.1 (3.6 GB)
        ├── wavs (13,100 .wavs)
        │   ├── LJ001-0001.wav
        │   └── ...
        ├── metadata.csv
        ├── README
        ├── train.txt
        ├── valid.txt
        └── test.txt

    """

    URL = "https://data.keithito.com/data/speech/LJSpeech-1.1.tar.bz2"

    # Train, valid, test splits files are downloaded from:
    # https://huggingface.co/datasets/flexthink/ljspeech/resolve/main/train.txt?download=true
    # https://huggingface.co/datasets/flexthink/ljspeech/resolve/main/valid.txt?download=true
    # https://huggingface.co/datasets/flexthink/ljspeech/resolve/main/test.txt?download=true

    DURATION = 86117.08  # Dataset duration (s), 24 hours

    def __init__(
        self, 
        root: str = None, 
        split: Literal["train", "valid" "test"] = "train",
        sr: float = 22050,  # Sampling rate
        crop: None | callable = StartCrop(clip_duration=10.),
        transform: None | callable = Mono(),
        target_transform: None | callable = None,
    ) -> None:
    
        self.root = root
        self.split = split
        self.sr = sr
        self.crop = crop
        self.transform = transform
        self.target_transform = target_transform

        if not Path(self.root).exists():
            raise Exception(f"{self.root} does not exist. Please download the dataset from {LJSpeech.URL}")

        self.meta_dict = self.load_meta()

    def __getitem__(self, index: int) -> dict:

        audio_path = self.meta_dict["audio_path"][index]
        caption = self.meta_dict["caption"][index]

        full_data = {
            "dataset_name": "LJSpeech",
            "audio_path": audio_path,
        }

        # Load audio data
        audio_data = self.load_audio_data(path=audio_path)
        full_data.update(audio_data)

        # Load target data
        target_data = self.load_target_data(caption=caption)
        full_data.update(target_data)

        return full_data

    def __len__(self) -> int:

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def load_meta(self) -> dict:
        r"""Load metadata of the GTZAN dataset.
        """

        # Load split file
        split_path = Path(self.root, "{}.txt".format(self.split))
        df = pd.read_csv(split_path, header=None)
        split_names = df[0].values

        # Load csv file
        csv_path = Path(self.root, "metadata.csv")
        df = pd.read_csv(csv_path, sep="|", header=None)
        audio_names = df[0].values
        captions = df[1].values

        # Get split indexes
        idxes = []
        for i in range(len(audio_names)):
            if audio_names[i] in split_names:
                idxes.append(i)

        audio_names = audio_names[idxes]
        captions = captions[idxes]
        audio_paths = [str(Path(self.root, "wavs", "{}.wav".format(name))) for name in audio_names]

        meta_dict = {
            "audio_name": audio_names,
            "audio_path": audio_paths,
            "caption": captions
        }

        return meta_dict

    def load_audio_data(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)

        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            clip_duration = audio_duration

        # Load a clip
        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )  # shape: (channels, audio_samples)

        # Transform audio
        if self.transform is not None:
            audio = call(transform=self.transform, x=audio)

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration
        }

        return data

    def load_target_data(self, caption: str) -> np.ndarray:

        target = caption

        # Transform target
        if self.target_transform:
            target = call(transform=self.target_transform, x=target)
            # target: (classes_num,)

        data = {
            "caption": caption,
            "target": target
        }

        return data


from torch.utils.data._utils.collate import collate, default_collate_fn_map
import torch


def collate_fn(batch: list):
    r"""The same as the default collate_fn in PyTorch. Except when input is a 
    list, then return a list."""

    def _collate_list_fn(batch, *, collate_fn_map=None):
        return batch

    collate_fn_map = default_collate_fn_map.copy()
    collate_fn_map.update({list: _collate_list_fn})

    return collate(batch, collate_fn_map=collate_fn_map)


from __future__ import annotations
from pathlib import Path
from typing import Optional

import librosa
import numpy as np
import pandas as pd
import os
from torch.utils.data import Dataset
from torch.utils.data._utils.collate import default_collate_fn_map

from audidata.io.audio import load
from audidata.io.crops import RandomCrop
from audidata.transforms.audio import Mono
from audidata.transforms.midi import MultiTrackPianoRoll
from audidata.io.midi import read_multi_track_midi
from audidata.collate.base import collate_list_fn


default_collate_fn_map.update({list: collate_list_fn})


class URMP(Dataset):
    r"""URMP [1] is a dataset comprises 44 simple multi-instrument classical music pieces assembled from coordinated but separately recorded performances of individual tracks. For each piece, the musical score is provided in MIDI format, the audio recordings of the individual tracks, the audio and video recording of the assembled mixture, and ground-truth annotation files including frame-level and note-level transcriptions are provided.

    [1] Bochen Li *, Xinzhao Liu *, Karthik Dinesh, Zhiyao Duan, Gaurav Sharma, "Creating a multi-track classical music performance dataset for multi-modal music analysis: Challenges, insights, and applications", IEEE Transactions on Multimedia, 2018. (* equal contribution)

    The dataset looks like:

        dataset_root (12 GB)
        ├── 01_Jupiter_vn_vc
        │   ├── AuMix_01_Jupiter_vn_vc.wav
        │   ├── AuSep_1_vn_01_Jupiter.wav
        │   ├── AuSep_2_vc_01_Jupiter.wav
        │   ├── F0s_1_vn_01_Jupiter.txt
        │   ├── F0s_2_vc_01_Jupiter.txt
        │   ├── Notes_1_vn_01_Jupiter.txt
        │   ├── Notes_2_vc_01_Jupiter.txt
        │   ├── Sco_01_Jupiter_vn_vc.mid
        ├── 02_Sonata_vn_vn
        ...
        ├── 43_Chorale_tpt_tpt_hn_tbn_tba
        ├── Supplementary_Files
    """

    url = "https://labsites.rochester.edu/air/projects/URMP.html"

    duration = 21355.7  # Dataset duration (s) - 5 hours 55 minutes 55.7 seconds

    def __init__(
        self,
        root: str,
        split: str = "train",
        sr: int = 44100,
        crop: Optional[callable] = RandomCrop(clip_duration=10.),
        transform: Optional[callable] = Mono(),
        target: bool = True,
        target_transform: Optional[callable] = MultiTrackPianoRoll(fps=100, pitches_num=128),
    ):
        self.root = Path(root)
        self.split = split
        self.sr = sr
        self.crop = crop
        self.transform = transform
        self.target = target
        self.target_transform = target_transform

        self.pieces = sorted([d for d in os.listdir(self.root) if d.startswith(("0", "1", "2", "3", "4"))]) # 01-44

    def __getitem__(self, index: int) -> dict:
        piece_dir = self.root / self.pieces[index]
        
        mix_file = next(piece_dir.glob("AuMix_*.wav"))
        stem_files = sorted(piece_dir.glob("AuSep_*.wav"))
        
        midi_file = next(piece_dir.glob("Sco_*.mid"))

        mix_audio = self.load_audio(mix_file)
        
        data = {
            "dataset_name": "URMP",
            "piece_name": self.pieces[index],
            "mix": mix_audio["audio"],
            "start_time": mix_audio["start_time"],
            "duration": mix_audio["duration"],
        }

        data["stems"] = []
        data["instruments"] = []
        for stem_file in stem_files:
            stem_audio = self.load_audio(stem_file, start_time=mix_audio["start_time"], duration=mix_audio["duration"])
            instrument = stem_file.stem.split("_")[2]
            data["stems"].append(stem_audio["audio"])
            data["instruments"].append(instrument)

        if self.target:
            midi_data = self.load_midi(midi_file, start_time=mix_audio["start_time"], clip_duration=mix_audio["duration"])
            data.update(midi_data)

        return data

    def __len__(self):
        return len(self.pieces)

    def load_audio(self, path: str, start_time: float = None, duration: float = None) -> dict:
        audio_duration = librosa.get_duration(path=path)
        
        if self.crop and start_time is None:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            clip_duration = duration

        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration if clip_duration else audio_duration
        }

        if self.transform is not None:
            data = self.transform(data)

        return data

    def load_midi(self, midi_path: str, start_time: float, clip_duration: float) -> dict:
        midi_data = read_multi_track_midi(midi_path=midi_path)

        data = {
            "tracks": [],
            "start_time": start_time,
            "clip_duration": clip_duration
        }

        for track in midi_data:
            track_data = {
                "program": track["program"],
                "is_drum": track["is_drum"],
                "notes": track["notes"],
                "start_time": start_time,
                "clip_duration": clip_duration
            }
            data["tracks"].append(track_data)

        return data


from __future__ import annotations
from pathlib import Path
from typing import Optional
import pandas as pd
import librosa
import numpy as np
from torch.utils.data import Dataset
from torch.utils.data._utils.collate import default_collate_fn_map

from audidata.io.audio import load
from audidata.io.crops import RandomCrop
from audidata.transforms.audio import Mono
from audidata.transforms.midi import PianoRoll
from audidata.io.midi import read_single_track_midi
from audidata.collate.base import collate_list_fn

default_collate_fn_map.update({list: collate_list_fn})

class GMD(Dataset):
    """The Groove MIDI Dataset (GMD) [1] is composed of 13.6 hours of aligned MIDI and (synthesized) audio of human-performed, tempo-aligned expressive drumming. The dataset contains 1,150 MIDI files and over 22,000 measures of drumming.
    
    [1] Jon Gillick, Adam Roberts, Jesse Engel, Douglas Eck, and David Bamman. "Learning to Groove with Inverse Sequence Transformations." International Conference on Machine Learning (ICML), 2019.
    """
    
    url = "https://magenta.tensorflow.org/datasets/groove"

    duration = 3944.09 # 10 hours, 51 minutes, 38.09 seconds

    def __init__(
        self,
        root: str,
        split: str = "train",
        sr: float = 16000,
        crop: Optional[callable] = RandomCrop(clip_duration=10., end_pad=9.9),
        transform: Optional[callable] = Mono(),
        target: bool = True,
        target_transform: Optional[callable] = PianoRoll(fps=100, pitches_num=128),
    ):
        self.root = Path(root)
        self.split = split
        self.sr = sr
        self.crop = crop
        self.target = target
        self.transform = transform
        self.target_transform = target_transform

        self.meta_dict = self.load_meta()

    def __getitem__(self, index: int) -> dict:
        audio_path = self.root / self.meta_dict["audio_filename"][index]
        midi_path = self.root / self.meta_dict["midi_filename"][index]

        full_data = {
            "dataset_name": "Groove MIDI Dataset",
            "audio_path": str(audio_path),
        }

        # Load audio
        audio_data = self.load_audio(path=audio_path)
        full_data.update(audio_data)

        # Load target
        if self.target:
            target_data = self.load_target(
                midi_path=midi_path,
                start_time=audio_data["start_time"],
                clip_duration=audio_data["duration"]
            )
            full_data.update(target_data)

        return full_data

    def __len__(self):
        return len(self.meta_dict["audio_filename"])

    def load_meta(self) -> dict:
        """Load metadata for the dataset."""
        csv_path = self.root / "info.csv"
        df = pd.read_csv(csv_path)
        df = df[df['split'] == self.split]
        
        meta_dict = {
            "midi_filename": df["midi_filename"].values,
            "audio_filename": df["audio_filename"].values,
            "duration": df["duration"].values,
            "style": df["style"].values,
            "bpm": df["bpm"].values,
            "beat_type": df["beat_type"].values,
            "time_signature": df["time_signature"].values
        }
        
        return meta_dict

    def load_audio(self, path: str) -> dict:
        audio_duration = librosa.get_duration(path=path)
        
        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            clip_duration = None

        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration if clip_duration else audio_duration
        }

        if self.transform is not None:
            data = self.transform(data)

        return data

    def load_target(self, midi_path: str, start_time: float, clip_duration: float) -> dict:
        notes, pedals = read_single_track_midi(
            midi_path=midi_path, 
            extend_pedal=False
        )

        data = {
            "note": notes,
            "pedal": pedals,
            "start_time": start_time,
            "clip_duration": clip_duration
        } 

        if self.target_transform:
            data = self.target_transform(data)

        return data

if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import soundfile
    from torch.utils.data import DataLoader

    root = "/root/groove"

    sr = 16000

    dataset = GMD(
        root=root,
        split="train",
        sr=sr,
        crop=RandomCrop(clip_duration=10., end_pad=9.9),
        target_transform=PianoRoll(fps=100, pitches_num=128),
    )

    dataloader = DataLoader(
        dataset=dataset, 
        batch_size=4, 
        num_workers=0, 
    )

    for data in dataloader:
        n = 0
        audio = data["audio"][n].cpu().numpy()
        frame_roll = data["frame_roll"][n].cpu().numpy()
        onset_roll = data["onset_roll"][n].cpu().numpy()
        offset_roll = data["offset_roll"][n].cpu().numpy()
        velocity_roll = data["velocity_roll"][n].cpu().numpy()
        break

    # ------ Visualize ------
    print("audio:", audio.shape)
    print("frame_roll:", frame_roll.shape)
    print("onset_roll:", frame_roll.shape)
    print("offset_roll:", frame_roll.shape)
    print("velocity_roll:", frame_roll.shape)

    # Write audio
    out_path = "out.wav"
    soundfile.write(file=out_path, data=audio.T, samplerate=sr)
    print("Write out audio to {}".format(out_path))

    # Mel spectrogram
    mel = librosa.feature.melspectrogram(y=audio[0], sr=sr, n_fft=2048, 
        hop_length=160, n_mels=229, fmin=0, fmax=8000)

    # Plot
    fig, axs = plt.subplots(5, 1, sharex=True, figsize=(20, 15))
    axs[0].matshow(np.log(mel), origin='lower', aspect='auto', cmap='jet')
    axs[1].matshow(frame_roll.T, origin='lower', aspect='auto', cmap='jet', vmin=0, vmax=1)
    axs[2].matshow(onset_roll.T, origin='lower', aspect='auto', cmap='jet', vmin=0, vmax=1)
    axs[3].matshow(offset_roll.T, origin='lower', aspect='auto', cmap='jet', vmin=0, vmax=1)
    axs[4].matshow(velocity_roll.T, origin='lower', aspect='auto', cmap='jet', vmin=0, vmax=1)
    fig_path = "out.pdf"
    plt.savefig(fig_path)
    print("Write out fig to {}".format(fig_path))


import random
from typing import Sized


class InfiniteSampler:
    def __init__(self, dataset: Sized) -> None:
        r"""Randomly sample indexes of a dataset without replacement. Execute
        this process infinitely.
        """

        self.indexes = list(range(len(dataset)))
        random.shuffle(self.indexes)  # self.indexes: [3, 7, 0, ...]
        self.p = 0  # pointer
        
    def __iter__(self) -> int:
        r"""Yield an index."""

        while True:

            if self.p == len(self.indexes):
                
                # Traversed all data. Reshuffle indexes. Reset pointer.
                random.shuffle(self.indexes)
                self.p = 0
                
            index = self.indexes[self.p]
            self.p += 1

            yield index


from __future__ import annotations

from audidata.datasets.audiocaps import AudioCaps
from audidata.datasets.clotho import Clotho
from audidata.datasets.clotho_aqa import ClothoAQA
from audidata.datasets.free_spoken_digit import FreeSpokenDigit
from audidata.datasets.gtzan import GTZAN
from audidata.datasets.librispeech import LibriSpeech
from audidata.datasets.ljspeech import LJSpeech
from audidata.datasets.maestro import MAESTRO
from audidata.datasets.musdb18hq import MUSDB18HQ
from audidata.datasets.shutterstock import Shutterstock
# from audidata.datasets.esc50 import ESC50
# from audidata.datasets.slakh2100 import Slakh2100
# from audidata.datasets.urmp import URMP
# from audidata.datasets.gmd import GMD
# from audidata.datasets.isophonics import Isophonics
from audidata.datasets.voicebank_demand import VoicebankDemand
# from audidata.datasets.magnatagatune import MagnaTagATune
from audidata.datasets.wavcaps import WavCaps



from audidata.tokenizers.base import (BaseTokenizer, SpecialTokenizer, 
	NameTokenizer, TimeTokenizer, PitchTokenizer, VelocityTokenizer, 
	ConcatTokenizer, ProgramTokenizer, DrumTokenizer, DictTokenizer)


import random
from typing import Optional

import librosa
import numpy as np
import torch
import torchaudio


'''
def load(
    path: str, 
    sr: int, 
    offset: float = 0.,  # Load start time (s)
    duration: Optional[float] = None,  # Load duration (s)
    mono: bool = False
) -> np.ndarray:
    r"""Load audio.

    Returns:
       audio: (channels, audio_samples) 

    Examples:
        >>> audio = load_audio(path="xx/yy.wav", sr=16000)
    """
    
    # Prepare arguments
    orig_sr = librosa.get_samplerate(path)

    start_sample = round(offset * orig_sr)

    if duration:
        samples = round(duration * orig_sr)
    else:
        samples = -1

    # Load audio
    audio, fs = torchaudio.load(
        path, 
        frame_offset=start_sample, 
        num_frames=samples
    )
    # (channels, audio_samples)

    # Resample. Faster than librosa
    audio = torchaudio.functional.resample(
        waveform=audio, 
        orig_freq=orig_sr, 
        new_freq=sr
    ).numpy()
    # shape: (channels, audio_samples)

    if duration:
        new_samples = round(duration * sr)
        audio = librosa.util.fix_length(data=audio, size=new_samples, axis=-1)

    if mono:
        audio = np.mean(audio, axis=0, keepdims=True)
    
    return audio
'''


def load(
    path: str, 
    sr: int, 
    offset: float = 0.,  # Load start time (s)
    duration: Optional[float] = None,  # Load duration (s)
    mono: bool = False
) -> np.ndarray:
    r"""Load audio.

    Returns:
       audio: (channels, audio_samples) 

    Examples:
        >>> audio = load_audio(path="xx/yy.wav", sr=16000)
    """
    
    # Load audio. librosa.load is faster than torchaudio.load
    audio, orig_sr = librosa.load(
        path, 
        sr=sr, 
        mono=mono, 
        offset=offset, 
        duration=duration
    )
    
    if audio.ndim == 1:
        audio = audio[None, :]
        # audio: (channels, audio_samples)

    # Resample. torchaudio's resample faster than librosa's resample
    audio = torchaudio.functional.resample(
        waveform=torch.Tensor(audio), 
        orig_freq=orig_sr, 
        new_freq=sr
    ).numpy()
    # audio: (channels, audio_samples)

    if duration:
        # Fix length to address 1) after resampling. 2) audio is shorter than duration
        audio = librosa.util.fix_length(
            data=audio, 
            size=round(duration * sr), 
            axis=-1
        )
        # audio: (channels, audio_samples)

    if mono:
        audio = np.mean(audio, axis=0, keepdims=True)
    
    return audio


from __future__ import annotations

import os
import re
from pathlib import Path

import librosa
import numpy as np
from audidata.io.audio import load
from audidata.io.crops import StartCrop
from audidata.transforms.audio import Mono
from audidata.transforms.onehot import OneHot
from audidata.utils import call
from torch.utils.data import Dataset
from typing_extensions import Literal


class GTZAN(Dataset):
    r"""GTZAN [1] is a music dataset containing 1,000 30-second audio clips. 
    The total duration is 8.3 hours. GTZAN includes 10 genres. All audio files 
    are mono sampled at 22,050 Hz. After decompression, the dataset size is 
    1.3 GB.

    [1] Tzanetakis, G., et al., Musical genre classification of audio signals. 2002

    The dataset looks like:

        gtzan (1.3 GB)
        └── genres
            ├── blues (100 files)
            ├── classical (100 files)
            ├── country (100 files)
            ├── disco (100 files)
            ├── hiphop (100 files)
            ├── jazz (100 files)
            ├── metal (100 files)
            ├── pop (100 files)
            ├── reggae (100 files)
            └── rock (100 files)
    """

    # The original webpage http://marsyas.info/index.html is no longer available anymore.
    URL = "https://huggingface.co/datasets/qiuqiangkong/gtzan/resolve/main/genres.tar.gz?download=true"

    DURATION = 30024.07  # Dataset duration (s), 8.3 hours

    LABELS = ["blues", "classical", "country", "disco", "hiphop", "jazz", 
        "metal", "pop", "reggae", "rock"]

    CLASSES_NUM = len(LABELS)
    LB_TO_IX = {lb: ix for ix, lb in enumerate(LABELS)}
    IX_TO_LB = {ix: lb for ix, lb in enumerate(LABELS)}

    def __init__(
        self, 
        root: str = None, 
        split: Literal["train", "test"] = "train",
        test_fold: int = 0,  # E.g., fold 0 is used for testing. Fold 1 - 9 are used for training.
        sr: float = 22050,  # Sampling rate
        crop: None | callable = StartCrop(clip_duration=30.),
        transform: None | callable = Mono(),
        target_transform: None | callable = OneHot(classes_num=CLASSES_NUM),
    ) -> None:
    
        self.root = root
        self.split = split
        self.test_fold = test_fold
        self.sr = sr
        self.crop = crop
        self.transform = transform
        self.target_transform = target_transform

        self.labels = GTZAN.LABELS
        self.lb_to_ix = GTZAN.LB_TO_IX
        self.ix_to_lb = GTZAN.IX_TO_LB

        if not Path(self.root).exists():
            raise Exception(f"{self.root} does not exist. Please download the dataset from {GTZAN.URL}")

        self.meta_dict = self.load_meta()

    def __getitem__(self, index: int) -> dict:

        audio_path = str(self.meta_dict["audio_path"][index])
        label = self.meta_dict["label"][index]

        full_data = {
            "dataset_name": "GTZAN",
            "audio_path": audio_path,
        }

        # Load audio data
        audio_data = self.load_audio_data(path=audio_path)
        full_data.update(audio_data)

        # Load target data
        target_data = self.load_target_data(label=label)
        full_data.update(target_data)

        return full_data

    def __len__(self) -> int:

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def load_meta(self) -> dict:
        r"""Load metadata of the GTZAN dataset.
        """

        meta_dict = {
            "audio_name": [],
            "audio_path": [],
            "label": [],
        }

        audios_dir = Path(self.root, "genres")

        for genre in self.labels:

            audio_names = sorted(os.listdir(Path(audios_dir, genre)))
            # E.g., len(audio_names) = 1000

            train_audio_names, test_audio_names = self.split_train_test(audio_names)
            # E.g., len(train_audio_names) = 900
            # E.g., len(test_audio_names) = 100

            if self.split == "train":
                filtered_audio_names = train_audio_names

            elif self.split == "test":
                filtered_audio_names = test_audio_names

            else:
                raise ValueError(self.split)

            for audio_name in filtered_audio_names:

                audio_path = str(Path(audios_dir, genre, audio_name))

                meta_dict["audio_name"].append(audio_name)
                meta_dict["audio_path"].append(audio_path)
                meta_dict["label"].append(genre)

        return meta_dict

    def split_train_test(self, audio_names: list) -> tuple[list, list]:

        train_audio_names = []
        test_audio_names = []

        test_ids = range(self.test_fold * 10, (self.test_fold + 1) * 10)
        # E.g., if test_fold = 3, then test_ids = [30, 31, 32, ..., 39]

        for audio_name in audio_names:

            audio_id = int(re.search(r'\d+', audio_name).group())
            # E.g., if audio_name is "blues.00037.au", then audio_id = 37

            if audio_id in test_ids:
                test_audio_names.append(audio_name)

            else:
                train_audio_names.append(audio_name)

        return train_audio_names, test_audio_names

    def load_audio_data(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)

        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            clip_duration = audio_duration

        # Load a clip
        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )
        # shape: (channels_num, audio_samples)

        # Transform audio
        if self.transform is not None:
            audio = call(transform=self.transform, x=audio)

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration
        }

        return data

    def load_target_data(self, label: str) -> dict:

        target = self.lb_to_ix[label]

        # Transform target
        if self.target_transform:
            target = call(transform=self.target_transform, x=target)

        data = {
            "label": label,
            "target": target  # shape: (classes_num,)
        }

        return data


from __future__ import annotations
import os
import re
import json
from pathlib import Path
from typing import Optional, Union

import librosa
import numpy as np
from torch.utils.data import Dataset

from audidata.io.audio import load
from audidata.io.crops import StartCrop
from audidata.transforms.audio import Mono
from audidata.utils import call


class WavCaps(Dataset):
    r"""WavCaps [1] is an audio caption dataset containing 402,958 audio 
    samples, each with 1 caption. The total duration is 7,545 hours. Most audio 
    samples are less than 10 seconds, with the longest audio lasting 12.39 
    hours. The audio files are mono and sampled at 32,000 Hz. After 
    decompression, the dataset size is 770 GB.

    [1] X. Mei, et al., WavCaps: A ChatGPT-Assisted Weakly-Labelled Audio 
    Captioning Dataset for Audio-Language Multimodal Research, 2023

    After decompression, the dataset looks like:

        wavcaps (131 GB)
        ├── Zip_files
        │   ├── AudioSet_SL (108,137 flac)
        │   ├── BBC_Sound_Effects (31,201 flac)
        │   ├── FreeSound (262,300 flac)
        │   └── SoundBible (1,320 flac)
        ├── Zip_files
        │   ├── AudioSet_SL
        │   │   └── as_final.json
        │   ├── BBC_Sound_Effects
        │   │   └── bbc_final.json
        │   ├── FreeSound
        │   │   └── fsd_final.json
        │   ├── SoundBible
        │   │   └── json_files_SoundBible_sb_final.json
        │   └── blacklist
        │       ├── blacklist_exclude_all_ac.json
        │       ├── blacklist_exclude_test_ac.json
        │       └── blacklist_exclude_ub8k_esc50_vggsound.json
        └── README.md
    """

    URL = "https://zenodo.org/records/3490684"

    DURATION = 27161470.93  # Dataset duration (s), 7,545 hours

    def __init__(
        self, 
        root: str = None, 
        sr: float = 32000,  # Sampling rate
        crop: None | callable = StartCrop(clip_duration=10.),
        transform: None | callable = None,
        target_transform: None | callable = None
    ) -> None:
    
        self.root = root
        self.sr = sr
        self.crop = crop
        self.transform = transform
        self.target_transform = target_transform

        if not Path(self.root).exists():
            raise Exception(f"{self.root} does not exist. Please download the dataset from {GTZAN.URL}")

        self.audios_dir = Path(self.root, "Zip_files")
        self.jsons_dir = Path(self.root, "json_files")
        paths = sorted(list(Path(self.jsons_dir).rglob("*.json")))

        json_paths = [path for path in paths if "blacklist" not in str(path)]
        blacklist_paths = [path for path in paths if "blacklist" in str(path)]

        self.meta_dict = self.load_meta(json_paths, blacklist_paths)

    def __getitem__(self, index: int) -> dict:

        caption = self.meta_dict["caption"][index]
        audio_path = self.meta_dict["audio_path"][index]
        subdataset = self.meta_dict["subdataset"][index]

        full_data = {
            "dataset_name": "WavCaps",
            "subdataset": subdataset,
            "audio_path": audio_path,
        }

        # Load audio data
        audio_data = self.load_audio_data(path=audio_path)
        full_data.update(audio_data)

        # Load target data
        target_data = self.load_target_data(caption=caption)
        full_data.update(target_data)

        return full_data

    def __len__(self) -> int:

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def load_meta(self, json_paths: list[str], blacklist_paths: list[str]) -> dict:

        black_names = self.get_black_names(blacklist_paths)
        # black_names: dict

        meta_dict = {"audio_name": [], "audio_path": [], "subdataset": [], "caption": []}

        for json_path in json_paths:
            subdataset = Path(json_path).parent.name
            with open(json_path, "r") as f:
                json_obj = json.load(f)
                for item in json_obj["data"]:
                    name = Path(item["id"]).stem
                    if name not in black_names.keys():

                        audio_name = "{}.flac".format(name)
                        audio_path = str(Path(self.audios_dir, subdataset, audio_name))

                        meta_dict["audio_name"].append(audio_name)
                        meta_dict["audio_path"].append(audio_path)
                        meta_dict["subdataset"].append(subdataset)
                        meta_dict["caption"].append(item["caption"])

        return meta_dict

    def get_black_names(self, blacklist_paths: list[str]) -> dict:

        black_names = {}

        for path in blacklist_paths:

            with open(path, "r") as f:

                json_obj = json.load(f)

                for subdataset in json_obj.keys():

                    subnames = json_obj[subdataset]
                    subnames = [Path(name).stem for name in subnames]

                    for name in subnames:

                        black_names[name] = True

        return black_names

    def load_audio_data(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)

        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            clip_duration = audio_duration

        # Load a clip
        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )  # shape: (channels_num, audio_samples)

        # Transform audio
        if self.transform is not None:
            audio = call(transform=self.transform, x=audio)

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration
        }

        return data

    def load_target_data(self, caption: str) -> dict:

        target = caption

        # Transform target
        if self.target_transform:
            target = call(transform=self.target_transform, x=target)

        data = {
            "caption": caption,
            "target": target,
        }

        return data


import copy
import math
import time

import numpy as np
from pretty_midi import ControlChange, Note, PrettyMIDI


class Pedal:
    def __init__(self, start: float, end: float):
        self.start = start
        self.end = end

    def __repr__(self):
        return "Pedal(start={:.4f}, end={:.4f})".format(self.start, self.end)


def read_single_track_midi(
    midi_path: str, 
    extend_pedal: bool
) -> tuple[list[Note], list[Pedal]]:
    r"""Read note and pedals from a single-track MIDI file.

    Returns:
        notes: list[Note]
        pedals: list[Pedal]
    """

    t1 = time.time()
    midi_data = PrettyMIDI(str(midi_path))
    print("c1", time.time() - t1)
    
    t1 = time.time()
    assert len(midi_data.instruments) == 1

    notes = midi_data.instruments[0].notes
    control_changes = midi_data.instruments[0].control_changes

    # Get pedals
    pedals = get_pedals(control_changes)
    print("c2", time.time() - t1)
    
    t1 = time.time()
    # Extend note offsets by pedal information
    if extend_pedal:
        notes = extend_offset_by_pedal(notes, pedals)
    print("c3", time.time() - t1)

    return notes, pedals

def read_multi_track_midi(midi_path: str) -> list[dict]:
    r"""Read notes from a multi-track MIDI file.
    Did not implement extend_pedal here. TODO: future work

    Returns:
        data: list[dict]
    """

    midi_data = PrettyMIDI(str(midi_path))
    
    data = []

    for instrument in midi_data.instruments:

        notes = instrument.notes
        control_changes = instrument.control_changes

        # Get pedals
        pedals = get_pedals(control_changes)

        data.append({
            "notes": notes,
            "pedals": pedals,
            "program": instrument.program,
            "is_drum": instrument.is_drum
        })

    return data


def get_pedals(control_changes: list[ControlChange]) -> list[Pedal]:
    r"""Get list of pedal events."""

    onset = None
    offset = None
    pedals = []

    control_changes.sort(key=lambda cc: cc.time)

    for cc in control_changes:

        if cc.number == 64:  # Sustain pedal

            if cc.value >= 64 and onset is None:
                onset = cc.time

            elif cc.value < 64 and onset is not None:
                offset = cc.time
                pedal = Pedal(start=onset, end=offset)
                pedals.append(pedal)
                onset = None
                offset = None

    if onset is not None and offset is None:
        offset = control_changes[-1].time
        pedal = Pedal(start=onset, end=offset)
        pedals.append(pedal)

    return pedals


def extend_offset_by_pedal(
    notes: list[Note], 
    pedals: list[Pedal]
) -> list[Note]:
    r"""Extend the note offset times to pedal offset times.

    Returns:
        new_notes: list[Note]
    """

    notes = copy.deepcopy(notes)
    pedals = copy.deepcopy(pedals)
    pitches_num = 128

    notes.sort(key=lambda note: note.end)

    notes_dict = {pitch: [] for pitch in range(pitches_num)}
    
    while len(pedals) > 0 and len(notes) > 0:

        pedal = pedals[0]  # Get the first pedal.

        while notes:

            note = notes[0]  # Get the first note.

            pitch = note.pitch

            if 0 <= note.end < pedal.start:
                notes_dict[pitch].append(note)
                notes.pop(0)

            elif pedal.start <= note.end < pedal.end:

                new_note = Note(
                    pitch=pitch, 
                    start=note.start, 
                    end=pedal.end, 
                    velocity=note.velocity
                )

                notes_dict[pitch].append(new_note)
                notes.pop(0)

            elif pedal.end <= note.end < math.inf:
                pedals.pop(0)
                break

            else:
                raise NotImplementedError 

    # Append remaining notes to dict
    for note in notes:
        notes_dict[note.pitch].append(note)

    # Note dict to note list
    new_notes = []
    for pitch in notes_dict.keys():
        new_notes.extend(notes_dict[pitch])
    
    new_notes.sort(key=lambda note: note.start)

    return new_notes


from __future__ import annotations

from pathlib import Path

import librosa
import pandas as pd
from audidata.io.audio import load
from audidata.io.crops import RandomCrop
from audidata.io.midi import read_single_track_midi, clip_notes
from audidata.transforms.audio import Mono
from audidata.transforms.midi import PianoRoll
from audidata.utils import call
from torch.utils.data import Dataset
from typing_extensions import Literal


class MAESTRO(Dataset):
    r"""MAESTRO [1] is a dataset containing 199 hours of audio from 1,276 files, 
    along with aligned MIDI files captured by a Yamaha Disklavier. All audio 
    files are stereo and sampled at 44,100 Hz. After decompression, the dataset 
    size is 131 GB.

    [1] C. Hawthorne, et al., Enabling Factorized Piano Music Modeling and 
    Generation with the MAESTRO Dataset, ICLR, 2019

    The dataset looks like:

        maestro-v3.0.0 (131 GB)
        ├── 2004 (132 songs, wav + flac + midi + tsv)
        ├── 2006 (115 songs, wav + flac + midi + tsv)
        ├── 2008 (147 songs, wav + flac + midi + tsv)
        ├── 2009 (125 songs, wav + flac + midi + tsv)
        ├── 2011 (163 songs, wav + flac + midi + tsv)
        ├── 2013 (127 songs, wav + flac + midi + tsv)
        ├── 2014 (105 songs, wav + flac + midi + tsv)
        ├── 2015 (129 songs, wav + flac + midi + tsv)
        ├── 2017 (140 songs, wav + flac + midi + tsv)
        ├── 2018 (93 songs, wav + flac + midi + tsv)
        ├── LICENSE
        ├── maestro-v3.0.0.csv
        ├── maestro-v3.0.0.json
        └── README
    """

    URL = "https://magenta.tensorflow.org/datasets/maestro"

    DURATION = 717232.49  # Dataset duration (s), 199 hours, including training, 
    # validation, and testing.

    def __init__(
        self, 
        root: str, 
        split: str = Literal["train", "validation", "test"],
        sr: float = 44100,
        crop: None | callable = RandomCrop(clip_duration=10., end_pad=9.9),
        transform: None | callable = Mono(),
        load_target: bool = True,
        extend_pedal: bool = True,
        target_transform: None | callable = PianoRoll(fps=100, pitches_num=128),
    ) -> None:

        self.root = root
        self.split = split
        self.sr = sr
        self.crop = crop
        self.load_target = load_target
        self.extend_pedal = extend_pedal
        self.transform = transform
        self.target_transform = target_transform

        if not Path(self.root).exists():
            raise Exception(f"{self.root} does not exist. Please download the dataset from {MAESTRO.URL}")

        meta_csv = Path(self.root, "maestro-v3.0.0.csv")

        self.meta_dict = self.load_meta(meta_csv)
        
    def __getitem__(self, index: int) -> dict:

        audio_path = self.meta_dict["audio_path"][index]
        midi_path = self.meta_dict["midi_path"][index]

        full_data = {
            "dataset_name": "MAESTRO",
            "audio_path": audio_path,
            "midi_path": midi_path
        }

        # Load audio data
        audio_data = self.load_audio_data(path=audio_path)
        full_data.update(audio_data)

        # Load target data
        if self.load_target:
            target_data = self.load_target_data(
                midi_path=midi_path,
                start_time=audio_data["start_time"],
                duration=audio_data["duration"]
            )
            full_data.update(target_data)

        return full_data

    def __len__(self) -> int:

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def load_meta(self, meta_csv: str) -> dict:
        r"""Load meta dict.
        """

        df = pd.read_csv(meta_csv, sep=',')

        indexes = df["split"].values == self.split

        audio_names = df["audio_filename"].values[indexes]
        midi_names = df["midi_filename"].values[indexes]
        durations = df["duration"].values[indexes]

        midi_paths = [str(Path(self.root, name)) for name in midi_names]
        audio_paths = [str(Path(self.root, name)) for name in audio_names]

        meta_dict = {
            "audio_name": audio_names,
            "audio_path": audio_paths,
            "midi_name": midi_names,
            "midi_path": midi_paths,
            "duration": durations
        }

        return meta_dict

    def load_audio_data(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)

        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            clip_duration = audio_duration

        # Load a clip
        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )
        # shape: (channels_num, audio_samples)

        # Transform audio
        if self.transform is not None:
            audio = call(transform=self.transform, x=audio)

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration
        }

        return data

    def load_target_data(
        self, 
        midi_path: str, 
        start_time: float, 
        duration: float
    ) -> dict:

        notes, pedals = read_single_track_midi(
            midi_path=midi_path, 
            extend_pedal=self.extend_pedal,
        )

        notes = clip_notes(notes, start_time, duration)
        pedals = clip_notes(pedals, start_time, duration)

        target = {
            "note": notes,
            "pedal": pedals,
            "start_time": start_time, 
            "duration": duration,
        }
        
        # Transform target
        if self.target_transform:
            target = call(transform=self.target_transform, x=target)

        return target


import random
from typing import NoReturn


class MUSDB18HQ_RandomSongSampler:
    def __init__(self, dataset) -> NoReturn:
        r"""Randomly sample indexes of different stems of a dataset without 
        replacement. Execute this process infinitely.
        """

        self.dataset = dataset
        self.stems = dataset.stems
        self.bg_stems = dataset.bg_stems
        self.target_stems = dataset.target_stems
        self.time_align = dataset.time_align
        
        self.idxes = self.get_indexes_dict()
        # E.g., {"bg": [3, 7, 0, ...], "target":, [4, 1, 9, ...]}
        
        self.ps = {key: 0 for key in self.idxes.keys()}  # pointers

    def __iter__(self) -> dict:
        r"""Yiled an index_dict."""

        while True:

            buffer = {}

            for key in self.idxes.keys():

                if self.ps[key] == len(self.idxes[key]):
                    # Traversed all data of a key. Reshuffle indexes. Reset pointer.
                    random.shuffle(self.idxes[key])
                    self.ps[key] = 0

                buffer[key] = self.idxes[key][self.ps[key]]
                self.ps[key] += 1

            index_dict = self.buffer_dict_to_indexes_dict(buffer)
            # E.g., {"vocals": 94, "drums": 94, "other": 35, "vocals": 35}
            
            yield index_dict

    def get_indexes_dict(self) -> dict:

        N = len(self.dataset)

        if self.time_align == "strict":
            idxes = {name: self.random_permutation(N) for name in ["song"]}

        elif self.time_align == "group":
            idxes = {name: self.random_permutation(N) for name in ["target", "bg"]}

        elif self.time_align == "random":
            idxes = {name: self.random_permutation(N) for name in self.stems}

        else:
            raise ValueError(self.time_align)

        return idxes

    def random_permutation(self, n) -> list:
        x = list(range(n))
        random.shuffle(x)
        return x

    def buffer_dict_to_indexes_dict(self, buffer: dict) -> dict:

        if self.time_align == "strict":
            idx_dict = {stem: buffer["song"] for stem in self.stems}

        elif self.time_align == "group":
            tar_idx_dict = {stem: buffer["target"] for stem in self.target_stems}
            bg_idx_dict = {stem: buffer["bg"] for stem in self.bg_stems}
            idx_dict = {**tar_idx_dict, **bg_idx_dict}  # Merge dict

        elif self.time_align == "random":
            idx_dict = buffer

        else:
            raise ValueError(self.time_align)

        return idx_dict


from audidata.samplers.infinite_sampler import InfiniteSampler
from audidata.samplers.pseudo_random_sampler import PseudoRandomSampler
from audidata.samplers.random_song_sampler import MUSDB18HQ_RandomSongSampler
from audidata.samplers.multi_datasets_batch_sampler import MultiDatasetsBatchSampler



from __future__ import annotations

from pathlib import Path

import librosa
import numpy as np
from audidata.io.audio import load
from audidata.io.crops import StartCrop
from audidata.transforms.audio import Mono
from audidata.utils import call
from torch.utils.data import Dataset
from typing_extensions import Literal


class FreeSpokenDigit(Dataset):
    r"""FreeSpokenDigit [1] is a speech dataset containing 3,000 recordings of 
    10 digits spoken by 6 speakers. The total duration of the dataset is 22 
    minutes. The durations of the audio clips range from 0.14 seconds to 2.28 
    seconds. All audio are mono sampled at 8 kHz. After decompression, the 
    dataset size is 58 MB.

    [1] https://github.com/Jakobovski/free-spoken-digit-dataset

    The dataset looks like:

        free-spoken-digit-dataset (22 MB)
        └── recordings (3,000 wavs)
            ├── 0_george_0.wav
            ├── 1_george_0.wav
            └── ...
    """

    URL = "https://github.com/Jakobovski/free-spoken-digit-dataset"

    DURATION = 1312.30  # Dataset duration (s), 22 minutes

    def __init__(
        self, 
        root: str = None, 
        split: Literal["train", "test"] = "train",
        sr: float = 8000,  # Sampling rate
        crop: None | callable = StartCrop(clip_duration=3.),
        transform: None | callable = Mono(),
    ) -> None:
    
        self.root = root
        self.split = split
        self.sr = sr
        self.crop = crop
        self.transform = transform

        if not Path(self.root).exists():
            raise Exception(f"{self.root} does not exist. Please download the dataset from {FreeSpokenDigit.URL}")

        self.speakers = self.get_speakers(split)

        self.meta_dict = self.load_meta()

    def __getitem__(self, index: int) -> dict:

        audio_path = self.meta_dict["audio_path"][index]
        label = self.meta_dict["label"][index]
        speaker = self.meta_dict["speaker"][index]

        full_data = {
            "dataset_name": "FreeSpokenDigit",
            "audio_path": audio_path,
        }

        # Load audio data
        audio_data = self.load_audio_data(path=audio_path)
        full_data.update(audio_data)

        # Load target data
        target_data = self.load_target_data(label=label, speaker=speaker)
        full_data.update(target_data)

        return full_data

    def __len__(self) -> int:

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def get_speakers(self, split: str) -> list[str]:

        if split == "train":
            speakers = ["jackson", "nicolas", "theo", "yweweler", "george"]

        elif split == "test":
            speakers = ["lucas"]

        else:
            raise ValueError(split)

        return speakers

    def load_meta(self) -> dict:
        r"""Load metadata of the dataset.
        """

        meta_dict = {
            "audio_name": [],
            "audio_path": [],
            "label": [],
            "speaker": []
        }
        
        audios_dir = Path(self.root, "recordings")

        audio_paths = sorted(list(Path(audios_dir).glob("*.wav")))

        for audio_path in audio_paths:

            audio_name = Path(audio_path).name
            label, speaker = audio_name.split("_")[0 : 2]

            if speaker in self.speakers:

                meta_dict["audio_name"].append(audio_name)
                meta_dict["audio_path"].append(str(audio_path))
                meta_dict["label"].append(label)
                meta_dict["speaker"].append(speaker)

        return meta_dict

    def load_audio_data(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)

        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            clip_duration = audio_duration

        # Load a clip
        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )  # shape: (channels_num, audio_samples)

        # Transform audio
        if self.transform is not None:
            audio = call(transform=self.transform, x=audio)

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration
        }

        return data

    def load_target_data(self, label: str, speaker: str) -> np.ndarray:

        caption = "speaker: {}, label: {}".format(speaker, label)

        data = {
            "label": label,
            "speaker": speaker,
            "caption": caption
        }

        return data


from __future__ import annotations
import random
import numpy as np

from torch.utils.data import Dataset


class MultiDatasetsBatchSampler:
    r"""Multiple datasets sampler. First, sample a dataset. Then, sample all 
    batch indexes from the sampled dataset. This class is useful when collating
    datasets with different keys in a batch. 
    """

    def __init__(
        self, 
        datasets: list[Dataset], 
        batch_size: int,
        weights: None | np.ndarray = None  # Probability of datasets to be sampled
    ) -> None:

        self.batch_size = batch_size
        self.datasets_num = len(datasets)
        
        # Probability of datasets to be sampled
        if weights is None:
            self.weights = [dataset.DURATION for dataset in datasets]
        else:
            self.weights = weights
            
        print("Dataset weights: {}".format(self.weights / np.sum(self.weights)))

        # Indexes of audio files of all datasets
        self.indexes_list = [list(range(len(dataset))) for dataset in datasets]

        for indexes in self.indexes_list:
            random.shuffle(indexes)

        self.datasets_lens = [len(indexes) for indexes in self.indexes_list]
        self.p_list = [0 for _ in range(self.datasets_num)]

    def __iter__(self) -> list[int]:
        r"""Sample a batch of indexes. First, sample a dataset. Then, sample all 
        batch indexes from the sampled dataset."""

        while True:

            # Dataset index
            m = random.choices(
                population=range(self.datasets_num), 
                weights=self.weights
            )[0]

            batch_indexes = []

            for _ in range(self.batch_size):

                if self.p_list[m] == self.datasets_lens[m]:
                
                    # Traversed all data. Reshuffle indexes. Reset pointer.    
                    random.shuffle(self.indexes_list[m])
                    self.p_list[m] = 0

                # Get index
                index = self.indexes_list[m][self.p_list[m]]
                self.p_list[m] += 1

                # Get absolute index
                abs_index = int(np.sum(self.datasets_lens[0 : m])) + index
                batch_indexes.append(abs_index)

            yield batch_indexes


from __future__ import annotations
from pathlib import Path
from typing import Optional

import yaml
import os
import librosa
import numpy as np
import pandas as pd
from torch.utils.data import Dataset
from torch.utils.data._utils.collate import default_collate_fn_map

from audidata.io.audio import load
from audidata.io.crops import RandomCrop
from audidata.transforms.audio import Mono
from audidata.transforms.midi import PianoRoll
from audidata.io.midi import read_single_track_midi, read_midi_beat
from audidata.collate.base import collate_list_fn


default_collate_fn_map.update({list: collate_list_fn})


class Slakh2100(Dataset):
    r"""Slakh2100 [1] is a multiple track MIDI-audio paired dataset containing
    145 hours of 2,100 audio audio files rendered by MIDI files. Audios are 
    sampled at 44,100 Hz. After decompression, the dataset is 101 GB.

    [1] E. Manilow, Cutting music source separation some Slakh: A dataset to 
    study the impact of training data quality and quantity, WASPAA, 2019

    After decompression, dataset looks like:

        dataset_root (131 GB)
        ├── train (1500 songs)
        │   ├── Track00001
        │   │   ├── all_src.mid
        │   │   ├── metadata.yaml
        │   │   ├── MIDI
        │   │   │   ├── S00.mid
        │   │   │   ├── S01.mid
        │   │   │   └── ...
        │   │   ├── mix.flac
        │   │   └── stems
        │   │       ├── S00.flac
        │   │       ├── S01.flac
        │   │       └── ...
        │   ├── Track00002
        │   └── ...
        ├── validation (375 songs)
        └── test (225 songs) 
    """

    url = "https://zenodo.org/records/4599666"

    duration = 521806.45  # Dataset duration (s), 145 hours, including training, 
    # validation, and testing.

    def __init__(
        self, 
        root: str, 
        split: str = "train",
        sr: float = 16000,
        crop: Optional[callable] = RandomCrop(clip_duration=10., end_pad=9.9),
        transform: Optional[callable] = Mono(),
        target: bool = True,
        extend_pedal: bool = True,
        target_transform: Optional[callable] = PianoRoll(fps=100, pitches_num=128),
    ):

        self.root = root
        self.split = split
        self.sr = sr
        self.crop = crop
        self.target = target
        self.extend_pedal = extend_pedal
        self.transform = transform
        self.target_transform = target_transform

        audios_dir = Path(self.root, self.split)
        self.meta_dict = {"audio_name": sorted(os.listdir(audios_dir))}

    def __getitem__(self, index: int) -> dict:

        prefix = Path(self.root, self.split, self.meta_dict["audio_name"][index])
        audio_path = Path(prefix, "mix.flac")
        meta_csv = Path(prefix, "metadata.yaml")
        mix_midi_path = Path(prefix, "all_src.mid")
        midis_dir = Path(prefix, "MIDI")

        full_data = {
            "dataset_name": "Slakh2100",
            "audio_path": str(audio_path),
        }

        # Load audio
        audio_data = self.load_audio(path=audio_path)
        full_data.update(audio_data)

        # Load target
        if self.target:
            target_data = self.load_target(
                meta_csv=meta_csv,
                mix_midi_path=mix_midi_path,
                midis_dir=midis_dir,
                start_time=audio_data["start_time"],
                clip_duration=audio_data["duration"]
            )
            full_data.update(target_data)

        return full_data

    def __len__(self):

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def load_meta(self, meta_csv: str) -> dict:
        r"""Load meta dict.
        """

        df = pd.read_csv(meta_csv, sep=',')

        indexes = df["split"].values == self.split

        meta_dict = {
            "midi_name": df["midi_filename"].values[indexes],
            "audio_name": df["audio_filename"].values[indexes],
            "duration": df["duration"].values[indexes]
        }

        return meta_dict

    def load_audio(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)
        
        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            clip_duration = None

        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )
        # shape: (channels, audio_samples)

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration if clip_duration else audio_duration
        }

        if self.transform is not None:
            data = self.transform(data)

        return data

    def load_target(
        self, 
        meta_csv: str,
        mix_midi_path: str,
        midis_dir: str,
        start_time: float, 
        clip_duration: float
    ) -> dict:
        
        with open(meta_csv, 'r') as f:
            meta = yaml.load(f, Loader=yaml.FullLoader)

        beats, downbeats = read_midi_beat(mix_midi_path)

        data = {
            "start_time": start_time,
            "clip_duration": clip_duration,
            "beat": beats,
            "downbeat": downbeats,
            "tracks": []
        }

        for stem_name, stem_data in meta["stems"].items():

            if not stem_data["midi_saved"]:
                continue

            inst_class = stem_data["inst_class"]
            is_drum = stem_data["is_drum"]
            plugin_name = stem_data["plugin_name"]
            program_num = stem_data["program_num"]

            midi_path = Path(midis_dir, "{}.mid".format(stem_name))

            notes, pedals = read_single_track_midi(
                midi_path=midi_path, 
                extend_pedal=self.extend_pedal
            )

            track = {
                "track_name": stem_name,
                "inst_class": inst_class,
                "is_drum": is_drum,
                "plugin_name": plugin_name,
                "program_num": program_num,
                "note": notes,
                "pedal": pedals,
            }

            data["tracks"].append(track)

        if self.target_transform:
            data = self.target_transform(data)

        return data


import numpy as np
import random
import librosa
import soundfile as sf


class ToMono:
    r"""Converts multi-channel audio to mono by averaging all channels.

    This transformation takes a multi-channel audio input and returns a mono audio
    by computing the mean across all channels.

    Args:
        None

    Returns:
        dict: A dictionary containing the mono audio under the key 'audio'.
    """

    def __init__(self):
        pass

    def __call__(self, data: dict) -> dict:
        data["audio"] = np.mean(data["audio"], axis=0, keepdims=True)
        return data


class Normalize:
    r"""Normalizes the audio to have a maximum absolute value of 1.

    This transformation scales the audio so that its maximum absolute value is 1.

    Args:
        None

    Returns:
        dict: A dictionary containing the normalized audio under the key 'audio'.
    """

    def __init__(self):
        pass

    def __call__(self, data: dict) -> dict:
        audio = data["audio"]
        audio = audio / np.max(np.abs(audio))
        data["audio"] = audio
        return data


class PitchShift:
    r"""Applies pitch shifting to the audio.

    This transformation shifts the pitch of the audio by a specified number of semitones.

    Args:
        n_steps (int): The number of semitones to shift the pitch.
        sr (int): The sampling rate of the audio.

    Returns:
        dict: A dictionary containing the pitch-shifted audio under the key 'audio'.
    """

    def __init__(self, n_steps: int, sr: int):
        self.n_steps = n_steps
        self.sr = sr

    def __call__(self, data: dict) -> dict:
        audio = data["audio"]
        shifted = np.apply_along_axis(lambda x: librosa.effects.pitch_shift(x, sr=self.sr, n_steps=self.n_steps), axis=-1, arr=audio)
        data["audio"] = shifted
        return data


class TimeStretch:
    r"""Applies time stretching to the audio.

    This transformation stretches or compresses the audio in time without changing its pitch.

    Args:
        rate (float): The rate at which to stretch the audio. Values > 1 speed up the audio,
                      while values < 1 slow it down.

    Returns:
        dict: A dictionary containing the time-stretched audio under the key 'audio'.
    """

    def __init__(self, rate: float):
        self.rate = rate

    def __call__(self, data: dict) -> dict:
        audio = data["audio"]
        stretched = np.apply_along_axis(lambda x: librosa.effects.time_stretch(x, rate=self.rate), axis=-1, arr=audio)
        data["audio"] = stretched
        return data


class AddGaussianNoise:
    r"""Adds Gaussian noise to the audio.

    This transformation adds random Gaussian noise to the audio signal.

    Args:
        noise_level (float): The standard deviation of the Gaussian noise to be added.

    Returns:
        dict: A dictionary containing the noisy audio under the key 'audio'.
    """

    def __init__(self, noise_level=0.005):
        self.noise_level = noise_level

    def __call__(self, data: dict) -> dict:
        audio = data["audio"]
        noise = np.random.normal(0, self.noise_level, audio.shape)
        noisy_audio = audio + noise
        data["audio"] = noisy_audio
        return data


class RandomGain:
    r"""Applies random gain to the audio.

    This transformation multiplies the audio by a random gain factor.

    Args:
        min_gain (float): The minimum gain factor.
        max_gain (float): The maximum gain factor.

    Returns:
        dict: A dictionary containing the gain-adjusted audio under the key 'audio'.
    """

    def __init__(self, min_gain=0.5, max_gain=1.5):
        self.min_gain = min_gain
        self.max_gain = max_gain

    def __call__(self, data: dict) -> dict:
        audio = data["audio"]
        gain = random.uniform(self.min_gain, self.max_gain)
        data["audio"] = audio * gain
        return data


class RandomImpulseReverb:
    r"""Applies random impulse reverb to the audio.

    This transformation creates a random impulse response and applies it to the audio
    to simulate reverb. The generated impulse response is normalized to prevent loudness changes.

    Args:
        rt60 (float): The RT60 (reverberation time) in seconds.
        room_scale (float): The scale of the simulated room.
        sr (int): The sampling rate of the audio.
        mix_ratio (float): The maximum ratio of reverb to original signal in the output.

    Returns:
        dict: A dictionary containing the reverberated audio under the key 'audio'.
    """

    def __init__(self, rt60=0.8, sr=22050, mix_ratio=0.5):
        self.rt60 = rt60
        self.sr = sr
        self.mix_ratio = mix_ratio

    def __call__(self, data: dict) -> dict:
        audio = data["audio"]
        reverb_audio = np.apply_along_axis(self.apply_random_impulse_reverb, axis=-1, arr=audio) # apply to each channel
        data["audio"] = reverb_audio
        return data

    def apply_random_impulse_reverb(self, audio):
        reverb_len = int(self.rt60 * self.sr)
        decay_constant = -np.log(0.001) / (self.rt60 * self.sr)
        time_points = np.arange(reverb_len) / self.sr
        decay = np.exp(-decay_constant * time_points) # exponential decay

        noise = np.random.randn(reverb_len)
        impulse_response = noise * decay
        impulse_response = impulse_response / np.linalg.norm(impulse_response) # normalize by power to prevent loudness changes

        reverb_audio = np.convolve(audio, impulse_response, mode='same')
        
        mixed_audio = (1 - self.mix_ratio) * audio + self.mix_ratio * reverb_audio
        return mixed_audio


def test_transformations(audio_path):
    r"""
    Testing function.
    """
    audio, sr = librosa.load(audio_path, sr=None, mono=False)
    if len(audio.shape) == 1:
        audio = audio[np.newaxis, :]  # Add channel dimension if mono, shape: (channels, audio_samples)
    original_data = {"audio": audio, "sr": sr}
    
    augmentations = [
        ToMono(),
        Normalize(),
        PitchShift(n_steps=2, sr=sr),
        TimeStretch(rate=0.5),
        AddGaussianNoise(noise_level=0.005),
        RandomGain(min_gain=0.5, max_gain=1.5),
        RandomImpulseReverb(rt60=0.1, sr=sr, mix_ratio=0.9)
    ]

    for i, aug in enumerate(augmentations):
        aug_data = aug(original_data.copy())
        aug_audio = aug_data["audio"]
        output_filename = f"augmented_{aug.__class__.__name__}.wav"

        sf.write(output_filename, aug_audio.T, sr)


if __name__ == "__main__":
    test_transformations("test.m4a")  # Provide path to an audio file for testing all transformations


from typing import Sized

import numpy as np


class PseudoRandomSampler:
    def __init__(self, dataset: Sized) -> None:
        r"""Pseudo random sampler. This sampler is usually used for evaluation.
        """

        rs = np.random.RandomState(1234)  # Random state
        self.indexes = rs.permutation(len(dataset))
        
    def __iter__(self) -> int:
        r"""Yield an index."""

        for index in self.indexes:
            yield index


from __future__ import annotations
import math
import librosa
import numpy as np
from pretty_midi import Note

from audidata.tokenizers.base import BaseTokenizer, DictTokenizer


class PianoRoll:
    r"""Convert the MIDI note and pedal events a full song into piano rolls of 
    a short clip. The rolls include frame roll, onset roll, offset roll, and
    velocity roll.
    """

    def __init__(
        self, 
        fps: int = 100, 
        pitches_num: int = 128, 
        soft_target: bool = False
    ):
        self.fps = fps
        self.pitches_num = pitches_num
        self.soft_target = soft_target

    def __call__(self, data: dict) -> dict:
        r"""Convert data dict to piano rolls."""

        notes = data["note"]
        pedals = data["pedal"]
        start_time = data["start_time"]
        clip_duration = data["duration"]

        clip_frames = round(self.fps * clip_duration) + 1

        # Rolls
        frame_roll = np.zeros((clip_frames, self.pitches_num), dtype="float32")
        onset_roll = np.zeros((clip_frames, self.pitches_num), dtype="float32")
        offset_roll = np.zeros((clip_frames, self.pitches_num), dtype="float32")
        velocity_roll = np.zeros((clip_frames, self.pitches_num), dtype="float32")

        if self.soft_target:
            soft_onset_roll = np.zeros((clip_frames, self.classes_num), dtype="float32")
            soft_offset_roll = np.zeros((clip_frames, self.classes_num), dtype="float32")

        clip_notes = []

        # Go through all notes
        for note in notes:

            onset_time = note.start - start_time
            offset_time = note.end - start_time
            pitch = note.pitch
            velocity = note.velocity

            if offset_time < 0:
                continue

            elif clip_duration < onset_time < math.inf:
                continue

            if offset_time == onset_time:
                offset_time = onset_time + (1. / fps)

            clip_note = Note(
                pitch=pitch, 
                start=onset_time, 
                end=offset_time, 
                velocity=velocity
            )
            clip_notes.append(clip_note)

            if offset_time < onset_time:
                raise ValueError("offset should not be smaller than onset!")
            
            # Update rolls
            elif onset_time < 0 and 0 <= offset_time <= clip_duration:

                offset_idx = round(offset_time * self.fps)
                offset_roll[offset_idx, pitch] = 1
                frame_roll[0 : offset_idx + 1, pitch] = 1

                if self.soft_target:
                    pass
                    # TODO but not necessary
                    # tmp = np.zeros(clip_frames)
                    # tmp[offset_idx] = 1
                    # delayed_frames = (offset_time * fps) % 1
                    # tmp = fractional_delay(tmp, delayed_frames)
                    # soft_offset_roll[:, pitch] += tmp
                    # from IPython import embed; embed(using=False); os._exit(0)

            elif onset_time < 0 and clip_duration < offset_time < math.inf:

                frame_roll[:, pitch] = 1

            elif 0 <= onset_time <= clip_duration and 0 <= offset_time <= clip_duration:

                onset_idx = round(onset_time * self.fps)
                offset_idx = round(offset_time * self.fps)
                onset_roll[onset_idx, pitch] = 1
                velocity_roll[onset_idx, pitch] = velocity / 128.0
                offset_roll[offset_idx, pitch] = 1
                frame_roll[onset_idx : offset_idx + 1, pitch] = 1

            elif 0 <= onset_time <= clip_duration and clip_duration < offset_time < math.inf:

                onset_idx = round(onset_time * self.fps)
                onset_roll[onset_idx, pitch] = 1
                velocity_roll[onset_idx, pitch] = velocity / 128.0
                frame_roll[onset_idx : , pitch] = 1

            else:
                raise NotImplementedError

        # Sort notes
        clip_notes.sort(key=lambda note: (note.start, note.pitch, note.end, note.velocity))

        data.update({
            "onset_roll": onset_roll,
            "offset_roll": offset_roll,
            "frame_roll": frame_roll,
            "velocity_roll": velocity_roll,
        })

        return data

class MultiTrackPianoRoll:
    r""" Will return a list of piano rolls for each track in the midi file.
    """
    def __init__(
        self,
        fps: int = 100,
        pitches_num: int = 128,
        soft_target: bool = False
    ):
        self.fps = fps
        self.pitches_num = pitches_num
        self.soft_target = soft_target

    def __call__(self, data: dict) -> dict:
        tracks = data["tracks"]
        start_time = data["start_time"]
        clip_duration = data["clip_duration"]

        clip_frames = round(self.fps * clip_duration) + 1

        # Generate rolls for each track
        track_rolls = []
        for track in tracks:
            track_roll = self.create_track_rolls(track, start_time, clip_duration, clip_frames)
            track_rolls.append(track_roll)

        data["track_rolls"] = track_rolls
        return data

    def create_track_rolls(self, track, start_time, clip_duration, clip_frames):
        notes = track["note"]
        
        frame_roll = np.zeros((clip_frames, self.pitches_num), dtype="float32")
        onset_roll = np.zeros((clip_frames, self.pitches_num), dtype="float32")
        offset_roll = np.zeros((clip_frames, self.pitches_num), dtype="float32")
        velocity_roll = np.zeros((clip_frames, self.pitches_num), dtype="float32")

        clip_notes = []

        for note in notes:
            onset_time = note.start - start_time
            offset_time = note.end - start_time
            pitch = note.pitch
            velocity = note.velocity

            if offset_time < 0 or onset_time >= clip_duration:
                continue

            if offset_time == onset_time:
                offset_time = onset_time + (1. / self.fps)

            clip_note = Note(
                pitch=pitch,
                start=max(0, onset_time),
                end=min(clip_duration, offset_time),
                velocity=velocity
            )
            clip_notes.append(clip_note)

            onset_idx = max(0, round(onset_time * self.fps))
            offset_idx = min(clip_frames - 1, round(offset_time * self.fps))

            if 0 <= onset_time < clip_duration:
                onset_roll[onset_idx, pitch] = 1
                velocity_roll[onset_idx, pitch] = velocity / 128.0

            if 0 <= offset_time <= clip_duration:
                offset_roll[offset_idx, pitch] = 1

            frame_roll[max(0, onset_idx):offset_idx + 1, pitch] = 1

        return {
            "frame_roll": frame_roll,
            "onset_roll": onset_roll,
            "offset_roll": offset_roll,
            "velocity_roll": velocity_roll,
            "clip_notes": clip_notes,
            "instrument": track.get("inst_class", "Unknown"),
            "is_drum": track.get("is_drum", False)
        }


from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="audidata", # Replace with your own username
    version="0.0.3",
    author="Qiuqiang Kong",
    author_email="qiuqiangkong@gmail.com",
    description="Audio dataloaders",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AudioFans/audidata",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["numpy", "torch", "torchaudio", "pretty_midi", "pandas", "pyyaml"],
    python_requires='>=3.5',
)


from __future__ import annotations

import os
from pathlib import Path

import librosa
import numpy as np
import pandas as pd
from audidata.io.audio import load
from audidata.io.crops import RandomCrop
from audidata.transforms.audio import Mono
from audidata.utils import call
from torch.utils.data import Dataset


class Shutterstock(Dataset):
    r"""Shutterstock is a private audio-text pairs music dataset consisting of 
    28,558 audio clips. The total duration is 1,250 hours. Each audio file has 
    1 caption, with captions containing 5 to 37 words. The durations of the 
    audio files range from 5 seconds to 1,415 seconds, with most audio samples 
    lasting between 1 and 3 minutes. The audio files are stereo and sampled at 
    48,000 Hz. After decompression, the dataset size is 409 GB.

    After decompression, the dataset looks like:

        dataset_root (409 GB)
        ├── flac (28,558 files)
        └── 28kdescriptions.csv

    """

    URL = "https://www.shutterstock.com/"

    DURATION = 4502350.60  # Dataset duration (s), 1,250 hours

    def __init__(
        self, 
        root: str = None, 
        sr: float = 48000,  # Sampling rate
        crop: None | callable = RandomCrop(clip_duration=10.),
        transform: None | callable = Mono(),
        target_transform: None | callable = None
    ) -> None:
    
        self.root = root
        self.sr = sr
        self.crop = crop
        self.transform = transform
        self.target_transform = target_transform

        self.meta_csv = Path(self.root, "28kdescriptions.csv")
        self.audios_dir = Path(self.root, "flac")

        if not Path(self.root).exists():
            raise Exception(f"{self.root} does not exist. Please download the dataset from {Shutterstock.URL}")

        self.meta_dict = self.load_meta(self.meta_csv, self.audios_dir)

    def __getitem__(self, index: int) -> dict:

        audio_path = self.meta_dict["audio_path"][index]
        caption = self.meta_dict["caption"][index]
        
        full_data = {
            "dataset_name": "Shutterstock",
            "audio_path": audio_path,
        }

        # Load audio data
        audio_data = self.load_audio_data(path=audio_path)
        full_data.update(audio_data)

        # Load target data
        target_data = self.load_target_data(caption=caption)
        full_data.update(target_data)

        return full_data

    def __len__(self) -> int:

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def load_meta(self, meta_csv: str, audios_dir: str) -> dict:

        audio_names_dict = {str(Path(name).stem): True for name in os.listdir(audios_dir)}

        df = pd.read_csv(meta_csv, sep=',')
        meta_dict = {"audio_name": [], "audio_path": [], "title": [], "caption": []}
        
        for n in range(len(df)):
            
            audio_name = str(df["idds"].values[n])

            if audio_name not in audio_names_dict.keys():
                continue

            audio_name = "{}.flac".format(audio_name)
            audio_path = str(Path(self.audios_dir, audio_name))

            meta_dict["audio_name"].append(audio_name)
            meta_dict["audio_path"].append(audio_path)
            meta_dict["title"].append(df["titles"].values[n])
            meta_dict["caption"].append(df["descri"].values[n])

        return meta_dict

    def load_audio_data(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)

        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            clip_duration = audio_duration

        # Load a clip
        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )  # shape: (channels_num, audio_samples)

        # Transform audio
        if self.transform is not None:
            audio = call(transform=self.transform, x=audio)

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration
        }

        return data

    def load_target_data(self, caption: str) -> np.ndarray:

        target = caption

        # Transform target
        if self.target_transform:
            target = call(transform=self.target_transform, x=target)
            # target: (classes_num,)

        data = {
            "caption": caption,
            "target": target
        }

        return data


from __future__ import annotations

import os
import re
from pathlib import Path

import pandas as pd
import librosa
import numpy as np
from audidata.io.audio import load
from audidata.io.crops import StartCrop
from audidata.transforms.audio import Mono
from audidata.transforms.onehot import OneHot
from audidata.utils import call
from torch.utils.data import Dataset
from typing_extensions import Literal


class LibriSpeech(Dataset):
    r"""LibriSpeech [1] is a speech dataset containing 292,367 English speech 
    files. The total duration is 1,000 hours. All audio files are mono and 
    sampled at 16,000 Hz. The durations of the audio files range from 0.83 s to 
    35.16 s, with 99.9% of the files having durations of less than 20 s. After 
    decompression, the dataset size is 60 GB.

    [1] Vassil Panayotov, et al., LibriSpeech: an ASR corpus based on public 
    domain audio books", ICASSP 2015

    The dataset looks like:

        librispeech (1.3 GB)
        ├── dev-clean (40 folders)
        │   ├── 1272 (3 folders)
        │   │   ├── 128104
        │   │   │   ├── 1272-128104-0000.flac
        │   │   │   ├── ...
        │   │   │   ├── 1272-128104-0014.flac
        │   │   │   └── 1272-128104.trans.txt
        │   │    ...
        │    ...
        ├── dev-other (33 folders)
        ├── test-clean (40 folders)
        ├── test-other (33 folders)
        ├── train-clean-100 (251 folders)
        ├── train-clean-360 (921 folders)
        ├── train-other-500 (1166 folders)
        ├── BOOKS.TXT
        ├── CHAPTERS.TXT
        ├── LICENSE.TXT
        ├── README.TXT
        └── SPEAKERS.TXT
    """

    # The original webpage http://marsyas.info/index.html is no longer available anymore.
    URL = "https://www.openslr.org/12"

    def __init__(
        self, 
        root: str = None, 
        split: str | list[str] = ["train-clean-100", "train-clean-360", "train-other-500"],
        sr: float = 16000,  # Sampling rate
        crop: None | callable = StartCrop(clip_duration=30.),
        transform: None | callable = Mono(),
        target_transform: None | callable = None,
    ) -> None:
    
        if isinstance(split, str):
            if split == "train":
                self.splits = ["train-clean-100", "train-clean-360", "train-other-500"]
            elif split == "dev":
                self.splits = ["dev-clean", "dev-other"]
            elif split == "test":
                self.splits = ["test-clean", "test-other"]
            else:
                raise ValueError

        elif isinstance(split, list):
            for name in split:
                assert name in ["train-clean-100", "train-clean-360", 
                    "train-other-500", "dev-clean", "dev-other", "test-clean", 
                    "test-other"]
            self.splits = split

        else:
            raise ValueError(split)


        self.root = root
        self.sr = sr
        self.crop = crop
        self.transform = transform
        self.target_transform = target_transform

        if not Path(self.root).exists():
            raise Exception(f"{self.root} does not exist. Please download the dataset from {LibriSpeech.URL}")

        self.meta_dict = self.load_meta()

    def __getitem__(self, index: int) -> dict:

        audio_path = str(self.meta_dict["audio_path"][index])
        caption = self.meta_dict["caption"][index]

        full_data = {
            "dataset_name": "LibriSpeech",
            "audio_path": audio_path,
        }

        # Load audio data
        audio_data = self.load_audio_data(path=audio_path)
        full_data.update(audio_data)

        # Load target data
        target_data = self.load_target_data(caption=caption)
        full_data.update(target_data)

        return full_data

    def __len__(self) -> int:

        audios_num = len(self.meta_dict["audio_name"])

        return audios_num

    def load_meta(self) -> dict:
        r"""Load metadata of the GTZAN dataset.
        """

        meta_dict = {
            "audio_name": [],
            "audio_path": [],
            "caption": [],
        }

        for split in self.splits:

            audios_dir = Path(self.root, split)

            for caption_path in Path(audios_dir).rglob('*trans.txt'):

                df = pd.read_csv(caption_path, sep="\t", header=None)
                data = df[0].str.split(" ", n=1, expand=True)

                audio_names = [str(name) for name in data[0].values]
                audio_paths = [str(Path(caption_path.parent, f"{name}.flac")) for name in audio_names]
                captions = [caption for caption in data[1].values]

                meta_dict["audio_name"].extend(audio_names)
                meta_dict["audio_path"].extend(audio_paths)
                meta_dict["caption"].extend(captions)
             
        return meta_dict

    def load_audio_data(self, path: str) -> dict:

        audio_duration = librosa.get_duration(path=path)
        
        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            clip_duration = audio_duration

        # Load a clip
        audio = load(
            path=path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )  # shape: (channels, audio_samples)

        # Transform audio
        if self.transform is not None:
            audio = call(transform=self.transform, x=audio)

        data = {
            "audio": audio, 
            "start_time": start_time,
            "duration": clip_duration
        }

        return data

    def load_target_data(self, caption: str) -> dict:

        target = caption

        # Transform target
        if self.target_transform:
            target = call(transform=self.target_transform, x=target)

        data = {
            "caption": caption,
            "target": target,
        }

        return data


from __future__ import annotations

from pathlib import Path

import librosa
from audidata.io.audio import load
from audidata.io.crops import RandomCrop
from audidata.transforms.audio import Mono, Normalize
from audidata.utils import call
from torch.utils.data import Dataset


class VoicebankDemand(Dataset):
    r"""Voicebank-Demand [1] contains 24,793 paired noisy and clean speech for 
    training speech enhancement algorithms. The total duration is 20 hours. 
    Audio files are mono are sampled at 16,000 Hz. After decompression, the 
    dataset is 2.5 GB.

    [1] Thiemann, J., et al. The diverse environments multi-channel acoustic 
    noise database (demand): A database of multichannel environmental noise 
    recordings. In Proceedings of Meetings on Acoustics, 2013

    The dataset looks like:

        voicebank-demand (2.5 GB)
        ├── clean_trainset_wav
        │    └── ... (11572 wavs)
        ├── noisy_trainset_wav
        │    └── ... (11572 wavs)
        ├── clean_testset_wav
        │    └── ... (824 wavs)
        └── noisy_testset_wav
             └── ... (824 wavs)
    """

    URL = "https://datashare.ed.ac.uk/handle/10283/2791"

    DURATION = 71761.32  # Dataset duration (s), 20 hours, including training 
    # and testing.

    def __init__(
        self, 
        root: str, 
        split: str = "train",
        sr: float = 16000,
        crop: None | callable = RandomCrop(clip_duration=2., end_pad=0.),
        stem_transform: None | callable = [Mono(), Normalize()],
    ) -> None:

        self.root = root
        self.split = split
        self.sr = sr
        self.crop = crop
        self.stem_transform = stem_transform

        if not Path(self.root).exists():
            raise Exception(f"{self.root} does not exist. Please download the dataset from {VoicebankDemand.URL}")

        self.meta_dict = self.load_metadata()
        
    def __getitem__(self, index: int) -> dict:

        # Audio paths
        mixture_audio_path = self.meta_dict["mixture_audio_path"][index]
        clean_audio_path = self.meta_dict["clean_audio_path"][index]

        full_data = {
            "dataset_name": "VoicebankDemand",
            "clean_audio_path": clean_audio_path,
            "mixture_audio_path": mixture_audio_path
        }

        # Load audio data
        audio_data = self.load_audio_data(mixture_audio_path, clean_audio_path)
        
        if self.stem_transform is not None:
            audio_data["target"] = call(self.stem_transform, audio_data["target"])
            audio_data["background"] = call(self.stem_transform, audio_data["background"])
            
        audio_data["mixture"] = audio_data["target"] + audio_data["background"]
        full_data.update(audio_data)
        
        return full_data

    def __len__(self):

        audios_num = len(self.meta_dict["clean_audio_path"])

        return audios_num

    def load_metadata(self) -> dict:
        r"""Load meta dict.
        """

        clean_audios_dir = Path(self.root, "clean_{}set_wav".format(self.split))
        clean_audio_paths = [str(path) for path in sorted(list(clean_audios_dir.glob("*.wav")))]
        
        mixture_audios_dir = Path(self.root, "noisy_{}set_wav".format(self.split))
        mixture_audio_paths = [str(Path(mixture_audios_dir, Path(path).name)) for path in clean_audio_paths]

        meta_dict = {
            "clean_audio_path": clean_audio_paths,
            "mixture_audio_path": mixture_audio_paths
        }

        return meta_dict

    def load_audio_data(self, mixture_audio_path: str, clean_audio_path: str) -> dict:

        audio_duration = librosa.get_duration(path=mixture_audio_path)

        if self.crop:
            start_time, clip_duration = self.crop(audio_duration=audio_duration)
        else:
            start_time = 0.
            clip_duration = audio_duration

        # Load mixture
        mixture = load(
            path=mixture_audio_path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )  # shape: (channels_num, audio_samples)

        # Load target
        target = load(
            path=clean_audio_path, 
            sr=self.sr, 
            offset=start_time, 
            duration=clip_duration
        )  # shape: (channels_num, audio_samples)

        background = mixture - target

        data = {
            "target": target, 
            "background": background,
            "start_time": start_time,
            "duration": clip_duration
        }

        return data